import re
from enum import Enum
from typing import Match

from hatsudenki.packages.expression.base import BaseExpression


class ExpressionMode(Enum):
    Default = 'ConditionExpression'
    Find = 'KeyConditionExpression'
    Filter = 'FilterExpression'


class KeyConditionExpression(BaseExpression):
    ParameterLabel = 'KeyConditionExpression'
    ValuePrefix = 'key'

    FilterWordMap = {
        'gt': 'greater',
        'lt': 'less',
        'gte': 'greater_than',
        'lte': 'less_than',
        'beginsWith': 'begins_with',
        'between': 'between'
    }

    FuncArgumentType = {
        'between': 'open',
    }

    def __init__(self):
        super().__init__()
        self.operations = ''
        self.val_num = 0

    def get_operation_by_word(self, word: str, *args, **kwargs):
        op = getattr(self, self.__class__.FilterWordMap.get(word, word), None)
        if op is None:
            Exception(f'invalid filter operation word. {word}')
        op(*args, **kwargs)

    def parse_key_value(self, key: str, value):
        ks = key.split('__')
        if len(ks) >= 2:
            op_func_name = self.__class__.FilterWordMap.get(ks[1], None)
            if op_func_name is None:
                Exception(f'invalid filter operation word. {ks[1]}')
            op_func = getattr(self, op_func_name)
            arg_type = self.__class__.FuncArgumentType.get(op_func_name)
            if arg_type == 'open':
                op_func(ks[0], *value)
            else:
                op_func(ks[0], value)
        else:
            self.equal(key, value)

    def is_empty(self):
        return len(self.operations) is 0

    def _make_func(self, op: str, key: str):
        k = self._register_key(key)
        self.operations += f'{op}({k})'

    def attribute_exists(self, key: str):
        self._make_func('attribute_exists', key)

    def attribute_not_exists(self, key: str):
        self._make_func('attribute_not_exists', key)

    def _make_comp(self, op: str, key: str, val: dict, raw=False):
        k = self._register_key(key)
        v = self._register_value(val, raw)
        op = f'{k} {op} {v}'
        self.operations += op

    def equal(self, key: str, val: any, raw=False):
        self._make_comp('=', key, val, raw)

    def greater(self, key: str, val: dict, raw=False):
        self._make_comp('>', key, val, raw)

    def greater_than(self, key: str, val: dict, raw=False):
        self._make_comp('>=', key, val, raw)

    def less(self, key: str, val: dict, raw=False):
        self._make_comp('<', key, val, raw)

    def less_than(self, key: str, val: dict, raw=False):
        self._make_comp('<=', key, val, raw)

    def between(self, key: str, lo: any, hi: any, raw=False):
        k = self._register_key(key)
        l = self._register_value(lo, raw)
        h = self._register_value(hi, raw)
        self.operations += f'{k} BETWEEN {l} AND {h}'

    def begins_with(self, key: str, val: str, raw=False):
        k = self._register_key(key)
        v = self._register_value(val, raw)
        self.operations += f'begins_with({k} ,  {v})'

    def op_and(self):
        if self.is_empty():
            return self
        self.operations = f'{self.operations} AND '
        return self

    def op_or(self):
        self.operations = f'{self.operations} OR '
        return self

    def __enter__(self):
        self.operations += '('

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.operations += ')'

    @property
    def expression(self):
        return self.operations

    @classmethod
    def dump(cls, data: dict):
        d = data[cls.ParameterLabel]
        k = data[cls.ParameterAttributeName.Names.value]
        v = data[cls.ParameterAttributeName.Values.value]

        def _(m: Match):
            return k[m.group()]

        def __(m: Match):
            return str(v[m.group()])

        p = re.sub(r'#key_\w+', _, d)
        return re.sub(r':key_\w+', __, p)


class ConditionExpression(KeyConditionExpression):
    ParameterLabel = 'ConditionExpression'
    ValuePrefix = 'get'

    FilterWordMap = {
        'in': 'in_values',
        'not': 'not_equal',
        **KeyConditionExpression.FilterWordMap
    }

    def in_values(self, key: str, val: list, raw=False):
        k = self._register_key(key)
        s = []
        for v in val:
            s.append(self._register_value(v))
        sa = ', '.join(s)
        op = f'{k} IN ({sa})'
        self.operations += op

    def not_equal(self, key: str, val: dict, raw=False):
        self._make_comp('<>', key, val, raw)


class FilterConditionExpression(ConditionExpression):
    ParameterLabel = 'FilterExpression'
    ValuePrefix = 'filter'



from __future__ import annotations

import ntpath
import re
import unicodedata
from pathlib import Path
from typing import Optional

import openpyxl
from openpyxl.cell import Cell
from openpyxl.worksheet.worksheet import Worksheet

from hatsudenki.packages.command.master.compare.schema import Schema


class ExcelSchemaReader(object):
    def __init__(self, base_path: Path):
        self.base_path = base_path
        self._data = {}

    def setup(self) -> ExcelSchemaReader:
        print(f'loading...xlsx files...')
        return self

    def schema(self) -> Schema:
        return Schema([self._read_group(p) for p in self.base_path.glob('*.xlsx') if not p.name.startswith('~$')])

    def _read_group(self, path) -> Schema.Group:
        grop_name = ntpath.basename(unicodedata.normalize('NFC', str(path))).replace('.xlsx', '')
        book = openpyxl.load_workbook(path)
        print(f'    loading...{grop_name}')
        return Schema.Group(grop_name, [t for t in (self._read_table(s) for s in book.worksheets) if t is not None])

    def _read_table(self, sheet: Worksheet) -> Optional[Schema.Group.Table]:
        if self._is_exclude_sheet(sheet):
            print(f'        skipping...{sheet.title}')
            return None
        print(f'        loading...{sheet.title}')
        table_name = sheet.title
        row = (c for c in (sheet.cell(1, i) for i in range(1, sheet.max_column)) if not self._is_exclude_cell(c))
        return Schema.Group.Table(table_name, [col for col in (self._read_column(c) for c in row) if col is not None])

    @staticmethod
    def _is_exclude_sheet(s: Worksheet):
        return s.title.startswith('★')

    def _read_column(self, cell: Cell) -> Optional[Schema.Group.Table.Column]:
        HYPERLINK_LEFT_REGEXP = re.compile('=HYPERLINK\(\"(([#!\.]?[^.]+)+)\", \"')
        cell_value = cell.value
        match_result = HYPERLINK_LEFT_REGEXP.match(cell_value)
        if match_result:
            delete_str = match_result.group()
            cell_value = cell_value.replace(delete_str, "")
            cell_value = cell_value.replace('")', "")
        return Schema.Group.Table.Column(cell_value)

    valid_colors = ['00CCFFCC', 'FFFCE4D6', '00FCE4D6']

    def _is_exclude_cell(self, c: Cell):
        if c.value is None:
            return True
        if c.fill.fgColor.value not in self.valid_colors:
            print(f"            skipping column {c.value} : {c.fill.fgColor.value}")
        return c.fill.fgColor.value not in self.valid_colors



import argparse
import asyncio
import importlib
import sys
from pathlib import Path
from traceback import print_exc

from hatsudenki.commands.base import BaseCommand
from hatsudenki.packages.command.stdout.output import ToolOutput


def get_args():
    command_src = (p.name.replace('.py', '') for p in Path(__file__).with_name('commands').glob('*.py') if
                   not p.name.startswith('__'))
    parser = argparse.ArgumentParser(description='execute managed commands.')
    sub_parsers = parser.add_subparsers()

    for com in command_src:
        if com == 'base':
            continue
        mod = importlib.import_module('hatsudenki.commands.' + com)
        if not hasattr(mod, 'Command'):
            print(f'Command is not defined in {com}.py')
            continue
        sub_parser = sub_parsers.add_parser(com, description=getattr(mod.Command, 'help', 'no description.'))
        com_ins: BaseCommand = mod.Command()
        sub_parser.add_argument('-q', '--quiet', action='store_true', help='no output log')
        com_ins.add_arguments(sub_parser)
        sub_parser.set_defaults(handler=com_ins.handle)

    return parser


def cli():
    # 引数の処理
    parser = get_args()
    args = parser.parse_args()
    if hasattr(args, 'handler'):
        s = vars(args)

        # quietがセットされていない場合はToolのロガーをセットアップする
        if not s.get('quiet'):
            ToolOutput.setup()

        acv = asyncio.get_event_loop()
        try:
            # commandの実行
            acv.run_until_complete(args.handler(*sys.argv, **s))
        except:
            print_exc()
    else:
        # helpの表示（一応）
        parser.print_help()


if __name__ == '__main__':
    cli()



from os import PathLike
from pathlib import Path
from typing import Dict

from hatsudenki.packages.command.loader.base import LoaderBase
from hatsudenki.packages.command.master.importer.yaml import MasterDataYaml
from hatsudenki.packages.command.master.loader import MasterTableLoader
from hatsudenki.packages.command.stdout.output import ToolOutput


class MasterDataYamlLoader(LoaderBase[MasterDataYaml]):

    def __init__(self, base_path: PathLike, loader: MasterTableLoader):
        super().__init__(base_path)
        self.ext = '.yml'
        self.master_loader = loader
        self.ref_table_name: Dict[str, MasterDataYaml] = {}

    def _load(self, path: Path):
        tbl_name = path.name.replace('.yml', '')
        ToolOutput.anchor(f'"{path}:1" をロード')
        table = self.master_loader.get_by_table_name(tbl_name)
        if table is None:
            ToolOutput.pop(f'みつからない {tbl_name}')
            return None

        a = MasterDataYaml(self.base_path, path, table, self)
        ToolOutput.pop('OK')
        return a

    def setup(self, dir_name=''):
        super().setup(dir_name)

        self.ref_table_name = {y.table_name: y for k, y in self.iter() if y is not None}



from enum import Enum


class BaseExpression(object):
    ParameterLabel = ''
    ValuePrefix = ''

    class ParameterAttributeName(Enum):
        Names = 'ExpressionAttributeNames'
        Values = 'ExpressionAttributeValues'

    def __init__(self):
        self.label = self.__class__.ParameterLabel
        self.prefix = self.__class__.ValuePrefix
        self.names = {}
        self._rev_names = {}
        self.values = {}
        self._rev_vals = {}
        self.val_num = 0
        self.key_num = 0

    @staticmethod
    def _put_dict_scalar(d: dict, k: str, val: any):
        if val is None:
            return

        t = type(val)
        if t in [str, list] and len(val) is 0:
            return
        if t is dict and not val:
            return

        d[k] = val

    @staticmethod
    def _put_dict_dict(d: dict, k: str, val: dict):
        if not val:
            return
        d[k] = val

    @staticmethod
    def _merge_dict_dict(d: dict, k: str, val: dict):
        if val is None:
            return
        if not val:
            return
        n: dict = d.get(k, None)
        if n is not None:
            n.update(val)
        else:
            d[k] = val

    @property
    def expression(self):
        raise NotImplementedError()

    def to_parameter(self, *merge_items: 'BaseExpression'):
        ret = {}

        val_label = self.ParameterAttributeName.Names.value
        name_label = self.ParameterAttributeName.Values.value

        self._put_dict_scalar(ret, str(self.label), self.expression)
        self._put_dict_dict(ret, val_label, self.names)
        self._put_dict_dict(ret, name_label, self.values)

        for m in merge_items:
            if m is None:
                continue
            self._put_dict_scalar(ret, str(m.label), m.expression)
            self._merge_dict_dict(ret, val_label, m.names)
            self._merge_dict_dict(ret, name_label, m.values)

        return ret

    @classmethod
    def merge(cls, *items: 'BaseExpression'):
        i = len(items)
        if i is 0:
            return {}

        arg = [item for item in items if item is not None]
        if len(arg) is 1:
            return arg[0].to_parameter()
        return arg[0].to_parameter(*arg[1:])

    def _register_key(self, key: str):
        """
        キー名を追加
        :param key: 追加するキー名
        :return: プレースホルダー文字列
        """
        # TODO: DictMapの更新を行った際に同じ内容のキーが複数生成されてしまう。すでにあるやつは使い回すべき

        # .で分割しないとmap型の更新に対応できない
        sps = key.split('.')

        r = []
        for sp in sps:
            # 他のExpressionとかぶらないようにPrefixを付与しておく（一応
            k = f'#{self.prefix}_key__{self.key_num}'

            # 配列判定
            bracket_pos = sp.find('[')
            if bracket_pos >= 0:
                bracket_end = sp.index(']')
                idx = sp[bracket_pos + 1:bracket_end]
                sp = sp[0:bracket_pos]

                old = self._rev_names.get(sp)
                if old:
                    k = old
                else:
                    self.names[k] = sp
                    self._rev_names[sp] = k
                    self.key_num += 1
                r.append(f'{k}[{idx}]')
            else:
                old = self._rev_names.get(sp)
                if old:
                    k = old
                else:
                    self.names[k] = sp
                    self._rev_names[sp] = k
                    self.key_num += 1
                r.append(k)
        return '.'.join(r)

    def _register_value(self, value: any, raw=False):
        """
        値を追加
        :param value: 追加する値
        :param raw: すでにシリアライズされている値を直接代入する場合はTrue
        :return: プレースホルダー文字列
        """
        # 循環参照の回避
        # rawフラグが設定されていない場合は内部でprimalシリアライザを使用する
        from hatsudenki.packages.field import primal_serializer
        sv = value if raw else primal_serializer(value)
        if sv is None:
            # DynamoDBは値をNoneで保持できないのでスキップ
            # （Noneを返すとクライアント側がスキップするようになっている）
            return None

        # TODO: 値の方のクエリ圧縮はそもそも必要なのか。tupleだとListで死ぬのでjsonで判定している（ダサい）
        # hashed_val = ujson.dumps(sv)
        # already = self._rev_vals.get(hashed_val)
        # if already:
        #     k = already
        # else:
        #     k = f':{self.prefix}_value__{self.val_num}'
        #     self.values[k] = sv
        #     self._rev_vals[hashed_val] = k
        #     self.val_num += 1

        k = f':{self.prefix}_value__{self.val_num}'
        self.values[k] = sv
        self.val_num += 1

        return k

    def is_empty(self):
        return False



from asyncio import futures
from concurrent import futures
from logging import getLogger
from pathlib import Path
from typing import Dict

from hatsudenki.packages.master.model import MasterModelSolo

_logger = getLogger(__name__)


def _build_inner(kls, yaml_dir_path: Path, cache_dir_path: Path):
    kls._load_from_yaml(yaml_dir_path, cache_dir_path, force=True)
    print(f'{kls.__name__} OK!!')


class MasterTableManager(object):
    _all_tables: Dict[str, MasterModelSolo] = {}
    _base_path: Path = None
    _dump_dir: Path = None

    @classmethod
    def register(cls, tbl):
        tbl_name = tbl.Meta.table_name
        if tbl_name is None:
            return

        if tbl_name in cls._all_tables:
            return

        cls._all_tables[tbl_name] = tbl

    @classmethod
    def get_by_table_name(cls, table_name: str):
        return cls._all_tables[table_name]

    @classmethod
    def iter_masters(cls):
        return cls._all_tables.items()

    @classmethod
    def all_load_from_yaml(cls, yaml_dir_path: Path, cache_dir_path: Path, force=False):
        _logger.info(f'load master from local_yaml {yaml_dir_path} CACHE=[{cache_dir_path}]')
        for k, t in cls._all_tables.items():
            t._load_from_yaml(yaml_dir_path, cache_dir_path, force=force)

    @classmethod
    def setup(cls, yml_path: Path):
        cls._base_path = yml_path

    @classmethod
    def _build_all(cls, yaml_dir_path: Path, cache_dir_path: Path, worker=2):
        with futures.ProcessPoolExecutor(max_workers=worker) as executor:
            l = []
            for k, t in cls._all_tables.items():
                print(k)
                l.append(executor.submit(_build_inner, t, yaml_dir_path, cache_dir_path))
            _ = futures.as_completed(l)



from typing import Optional, List

from hatsudenki.packages.command.hatsudenki.field import field, HatsudenkiFieldBase


@field('string', 'str')
class StringField(HatsudenkiFieldBase):
    PythonStr = 'field.StringField'

    @property
    def default_value(self):
        d = super().default_value
        return f"'{d}'" if d is not None else None

    @property
    def gql_filters(self) -> Optional[List[str]]:
        return ['beginsWith']


@field('number', 'int')
class NumberField(HatsudenkiFieldBase):
    PythonStr = 'field.NumberField'

    @property
    def gql_filters(self) -> Optional[List[str]]:
        return ['gt', 'gte', 'lte', 'lt']


@field('set_number', 'set')
class NumberSetField(HatsudenkiFieldBase):
    PythonStr = 'field.NumberSetField'


@field('set_string', 'set')
class StringSetField(HatsudenkiFieldBase):
    PythonStr = 'field.StringSetField'


@field('bool', 'bool')
class BoolField(HatsudenkiFieldBase):
    TypeStr = 'bool'
    PythonStr = 'field.BoolField'


@field('binary', 'bytes')
class BinaryField(HatsudenkiFieldBase):
    PythonStr = 'field.BinaryField'

    @property
    def default_value(self):
        return None



from logging import getLogger, ERROR
from typing import Type, TypeVar, List, Dict, Tuple, Generator, Callable, Optional, AsyncGenerator

from hatsudenki.packages.client import HatsudenkiClient
from hatsudenki.packages.expression.condition import ConditionExpression, KeyConditionExpression, \
    FilterConditionExpression
from hatsudenki.packages.expression.update import UpdateExpression
from hatsudenki.packages.field import NumberField
from hatsudenki.packages.field.base import BaseHatsudenkiField
from hatsudenki.packages.field.extra import CreateDateField, UpdateDateField
from hatsudenki.packages.manager.date import DateManager
from hatsudenki.packages.table.define import TableType
from hatsudenki.packages.table.index import PrimaryIndex

T = TypeVar('T')

_logger = getLogger(__name__)


class SoloHatsudenkiTable(object):
    # テーブルメタ情報定義
    class Meta:
        label = 'default'
        is_root = True
        # テーブル名
        collection_name = ''
        primary_index: PrimaryIndex = None
        table_name = ''

    class Field:
        _v = NumberField()

    class LSIndex:
        pass

    class GSIndex:
        pass

    _attributes: Dict[str, BaseHatsudenkiField] = {}
    _serializer: Dict[str, Callable] = {}
    _deserializer: Dict[str, Callable] = {}
    _not_scalar_key: List[str] = []
    _hook_update_key: List[BaseHatsudenkiField] = []
    _hook_put_key: List[BaseHatsudenkiField] = []
    _hash_key_name: Optional[str] = None
    _range_key_name: Optional[str] = None
    _collection_name: Optional[str] = None

    @classmethod
    def get_table_type(cls):
        if cls.Meta.is_root:
            return TableType.RootTable
        return TableType.SingleSoloTable

    @classmethod
    def _is_allow_modify(cls, key: str):
        hash_key = cls.get_hash_key_name()
        if key == hash_key:
            # ハッシュキーは変更できない
            return False
        return True

    def __init__(self, **kwargs):
        self._update_keys: Dict[str, any] = {}
        self._v = 0
        self._is_new = True

    def _modify_mark(self, key):
        self._update_keys.setdefault(key, self.__dict__.get(key, None))

    def __init_subclass__(cls, **kwargs):
        from hatsudenki.packages.manager.table import TableManager
        super().__init_subclass__(**kwargs)
        # マネージャーに登録

        if not TableManager.register(cls):
            return

        # キー情報をオーバーライド
        cls._attributes = {}
        cls._serializer = {}
        cls._deserializer = {}
        cls._not_scalar_key = []
        cls._hash_key_name = None
        cls._range_key_name = None
        cls._collection_name = None

        cls._hook_update_key = []
        cls._hook_put_key = []
        for k in dir(cls.Field):
            v = getattr(cls.Field, k)
            if isinstance(v, BaseHatsudenkiField) is False:
                # Field型ではない
                continue
            v.name = k
            cls._attributes[k] = v
            if isinstance(v, CreateDateField):
                cls._hook_put_key.append(v)
            elif isinstance(v, UpdateDateField):
                cls._hook_update_key.append(v)

            cls._serializer[k] = v.serialize
            cls._deserializer[k] = v.deserialize
            if not v.IsScalar:
                cls._not_scalar_key.append(k)

        cls._hash_key_name = cls.Meta.primary_index.hash_key
        cls._range_key_name = cls.Meta.primary_index.range_key
        cls._collection_name = cls.Meta.collection_name

        if hasattr(cls.Meta, 'alias_key_type'):
            cls.Meta.alias_key_type.name = cls.Mata.alias_key_name

    def __getitem__(self, item):
        # self[xxx]で属性にアクセスできるようにしている
        return self.__dict__[item]

    def __setattr__(self, key: str, value, force=False):
        if force or key not in self.__class__._attributes:
            # 以下の場合は何もしない
            # 強制フラグがON
            # はじめてのセット
            # DB定義されたフィールドでない
            object.__setattr__(self, key, value)
            return

        is_first_set = not hasattr(self, key)
        prev = self[key] if not is_first_set else None

        # ハッシュキーもしくはレンジキーは書き換えることが出来ない
        is_allow = self.__class__._is_allow_modify(key)
        if not is_allow and prev is not None:
            raise Exception(f'can not modify hash_key os range_key {key}')
        # 変更されたキーに追加
        if is_allow:
            self.modify_mark(key)
        super().__setattr__(key, value)

    def modify_mark(self, key):
        prev = getattr(self, str(key), None)
        self._update_keys.setdefault(key, prev)

    @property
    def hash_key_name(self):
        return self.__class__.get_hash_key_name()

    @classmethod
    def get_hash_key_name(cls):
        p = cls.get_primary_index()
        return p.hash_key

    @classmethod
    def get_collection_name(cls):
        return cls._collection_name

    @classmethod
    def get_table_name(cls):
        """
        テーブル名を取得

        :return: テーブル名文字列
        """
        return cls.Meta.table_name

    @classmethod
    def get_primary_index(cls):
        return cls.Meta.primary_index

    @classmethod
    def get_indexes(cls):
        from hatsudenki.packages.manager.table import TableManager
        return TableManager.get_index(cls.get_collection_name())

    @classmethod
    def get_field_class(cls, key):
        return cls._attributes.get(key, None)

    @classmethod
    def get_hash_field_class(cls):
        return cls.get_field_class(cls.get_hash_key_name())

    @property
    def hash_value(self):
        """
        ハッシュキーの値を取得

        :return: ハッシュキーの値（型はモデルに依存）
        """
        return getattr(self, self.hash_key_name)

    @property
    def serialized_key(self):
        """
        インスタンスのシリアライズされたハッシュキーとレンジキーを取得

        :return: シリアライズされたキー情報を格納した連想配列
        """

        c = self.__class__
        hk = c._hash_key_name

        return {
            hk: c._serializer[c._hash_key_name](self[hk])
        }

    def force_set_key(self, key, val):
        self.__setattr__(key, val, True)

    @classmethod
    def deserialize(cls: Type[T], raw_dict: dict, prj: List[str] = None) -> T:
        """
        AWSレスポンスからインスタンスを生成する

        :param raw_dict: awsアイテムレスポンス
        :return: モデルインスタンス
        """
        ret = cls()
        _attr = cls._attributes
        _set = object.__setattr__
        _der = cls._deserializer

        for key, val in raw_dict.items():
            if key in _der:
                _set(ret, key, _der[key](val, ret))

        # ユーザー操作によって更新されたものではないので更新フラグはここで一旦すべて折ってきれいにする
        # DictMap関連のappendedフラグもここで折れる
        ret.flush()
        # 新たに作られたものではない
        ret._is_new = False

        if prj is not None:
            ret._set_projection(prj)

        return ret

    def flush(self):
        """
        更新キー情報をクリア

        :return: None
        """

        d = self.__dict__

        for k in self._not_scalar_key:
            d[k].flush()

        self._update_keys = {}
        self._is_new = False

    def serialize(self):
        """
        シリアライズされたアイテム情報を取得
        :return: シリアライズされたアイテム情報を格納した連想配列
        """
        a = self.__dict__
        ret = {}
        for k, v in self.__class__._serializer.items():
            vv = v(a[k])
            if vv is not None:
                ret[k] = vv

        return ret

    @classmethod
    def not_exist_condition(cls, cond: ConditionExpression = None):
        """
        存在を確認する条件式インスタンスを生成

        :return: 存在確認式が設定された条件式インスタンス
        """
        if cond is None:
            cond = ConditionExpression()
        cond.attribute_not_exists(cls.get_hash_key_name())
        return cond

    @classmethod
    def exist_condition(cls, cond: ConditionExpression = None):
        """
        存在を確認する条件式インスタンスを生成

        :return: 存在確認式が設定された条件式インスタンス
        """
        if cond is None:
            cond = ConditionExpression()
        cond.attribute_exists(cls.get_hash_key_name())
        return cond

    def _hook_put(self):
        for pk in self.__class__._hook_put_key:
            # self.force_set_key(pk.name, TableManager.resolve_date_now())
            setattr(self, pk.name, DateManager.get_now())

    async def put(self, overwrite=False, skip_hook=False):
        """
        新規作成
        overwriteがFalse且つアイテムが存在する場合は例外が発生する

        :param overwrite: 強制上書きを行うか
        :param skip_hook: フック処理をスキップするか
        :param with_retry: リトライ処理を行うか
        :param error_level: 失敗時のエラーレベル
        :return: awsレスポンス
        """
        if not skip_hook:
            self._hook_put()
        # 強制的にカウンタを初期値に戻す
        self.force_set_key('_v', self._v + 1)
        ser = self.serialize()
        c = self.not_exist_condition() if not overwrite else None

        res = await HatsudenkiClient.put_item(
            self.get_collection_name(),
            ser,
            c
        )

        self.flush()
        return res

    def _hook_update(self):
        for pk in self.__class__._hook_update_key:
            # self.force_set_key(pk.name, TableManager.resolve_date_now())
            setattr(self, pk.name, DateManager.get_now())

    def build_update_expression(self, upd: UpdateExpression):
        for update_key, prev_val in self._update_keys.items():
            if update_key == '_v':
                # _vは後で強制的に入れるのでここでは無視
                continue
            # フィールドクラス取得
            fc = self.__class__.get_field_class(update_key)
            if fc.is_empty(self[update_key]):
                # 値が空且つ変更が行われた＝値が削除された
                upd.remove(update_key)
                continue

            # フィールドがスカラー型ではない場合は更に潜ってチェック
            now_value = self[update_key]
            fc.build_update_expression(upd, now_value)

            # upd.set(update_key, fc.serialize(self[update_key]), raw=True)

    async def update(self, upsert=False, increment=True, skip_hook=False):
        """
        更新
        upsertが偽且つアイテムが存在しない場合は例外が発生する
        :param upsert: レコードが存在しない場合新規作成するか
        :param increment: アトミックカウンターを考慮するか
        :param skip_hook: フック処理をスキップするか
        :param with_retry: 失敗時にリトライを行うか
        :param error_level: 失敗時のエラーレベル
        :return: awsレスポンス
        """
        if len(self._update_keys) is 0:
            # 何も更新されていないのでスキップ
            _logger.debug('update key is nothing. skip update...')
            return

        if not skip_hook:
            self._hook_update()

        # 更新されたものだけ候補に入れる
        upd = UpdateExpression()

        self.build_update_expression(upd)

        cond = self.exist_condition() if not upsert else None

        if increment:
            # バージョンカウンタの操作
            upd.add('_v', 1)
            if not upsert:
                if cond is None:
                    cond = ConditionExpression()
                else:
                    cond.op_and()

                with cond:
                    cond.equal('_v', self._v)
                    cond.op_or()
                    # アイテムが存在しない時用
                    cond.attribute_not_exists('_v')

        keys = self.serialized_key

        res = await HatsudenkiClient.update_item(self.get_collection_name(), keys, upd, cond)

        self.flush()

        if increment:
            self._v += 1

        return res

    async def reget(self):
        c = await self.__class__.get(self.hash_value)
        self.copy(c)

    def copy(self, model):

        [self.force_set_key(k, model[k]) for k, v in self._attributes.items()]
        self.flush()

    @classmethod
    def get_serialized_key(cls, hash_val: any, range_val: any = None):
        """
        シリアライズされたハッシュキーとレンジキーを取得する

        :param hash_val: ハッシュキーの値（型はmodelに依存）
        :param range_val: レンジキーの値（型はmodelに依存）
        :return: シリアライズされたキー情報を格納した連想配列
        """
        cls._check_key(hash_val, range_val)

        hash_key = cls.get_hash_key_name()

        hc = cls.get_field_class(hash_key)
        ret = {
            hash_key: hc.serialize(hash_val)
        }

        return ret

    @classmethod
    def find_index(cls, hash_key: str, range_key: str = None):
        index = cls.get_indexes()
        if index.primary.check(hash_key, range_key):
            return index.primary

        if index.primary.hash_key == hash_key:
            for lsi in index.lsi:
                if lsi.check(hash_key, range_key):
                    return lsi

        for gsi in index.gsi:
            if gsi.check(hash_key, range_key):
                return gsi

        # childテーブルから参照した際に、Indexが見つからないのは正常な挙動
        # _logger.warning(f'{cls.get_table_name()} [{hash_key}, {range_key}] : INDEX NOT FOUND!!!')

        return None

    @classmethod
    async def get(cls, hash_val: any, range_val: any = None, prj_exp: List[str] = None):
        """
        ハッシュキーとレンジキー（あれば）を指定して一見取得
        :param hash_val: ハッシュキーの値
        :return: モデルインスタンス。アイテムが見つからない場合はNone
        """
        cls._check_key(hash_val, range_val)

        k = cls.get_serialized_key(hash_val, range_val)
        return await cls.get_raw(k, prj_exp)

    @classmethod
    async def get_raw(cls, raw_dict: dict, prj_exp: List[str] = None):
        r = await HatsudenkiClient.get_item(cls.get_collection_name(), raw_dict, prj_exp)
        if r is None:
            return None

        return cls.deserialize(r, prj_exp)

    @classmethod
    async def get_iter(cls: Type[T], hash_val: any, limit=0, prj_exp: List[str] = None) -> Generator[T, None, None]:
        query_dict = {
            cls.get_hash_key_name(): hash_val
        }
        return await cls.query_list(query_dict, limit, prj_exp)

    @classmethod
    def from_raw_dict_list(cls, raw_dict_list: List[dict]):
        """
        AWSレスポンスからインスタンスを生成する（配列版）

        :param raw_dict_list: awsアイテムリストレスポンス
        :return: モデルインスタンス
        """
        return map(cls.deserialize, raw_dict_list)

    @classmethod
    async def query_list(cls: Type[T],
                         query_dict: dict, limit=0, prj_exp: List[str] = None,
                         filter_dict: dict = None) -> Generator[T, None, None]:
        kc, idx = cls.query_parse(query_dict)
        index_name = idx.name if idx is not None else None
        fc = None
        if filter_dict:
            fc = cls.filter_parse(filter_dict)

        res = await HatsudenkiClient.query(table_name=cls.get_collection_name(), key_cond=kc,
                                           use_index_name=index_name, limit=limit, prj=prj_exp, filter_cond=fc)
        return cls.from_raw_dict_list(res)

    # @classmethod
    # async def paginator(cls: Type[T], query_dict: dict, page_size=100) -> Generator[T, None, None]:

    @classmethod
    async def query(cls: Type[T], query_dict: dict, prj_exp: List[str] = None, filter_dict: dict = None) -> T:
        l = await cls.query_list(query_dict, 1, prj_exp, filter_dict)
        return next(l, None)

    @classmethod
    async def query_by_cursor(cls: Type[T], cursor: str) -> T:
        return await cls.get(cursor)

    @classmethod
    async def query_list_by_cursor(cls: Type[T], cursor: str) -> List[T]:
        raise Exception('invalid operation')

    @classmethod
    async def scan_iter(cls: Type[T], filter_dict: dict = None, limit=20, prj: List[str] = None) -> \
            List[T]:
        fc = None
        if filter_dict:
            fc = cls.filter_parse(filter_dict)
        l = await HatsudenkiClient.scan(
            table_name=cls.get_collection_name(), filter_cond=fc, limit=limit, prj=prj)

        return cls.from_raw_dict_list(l)

    @classmethod
    async def scan_generator(cls: Type[T],
                             filter_dict: dict = None, tick=50, prj: List[str] = None) -> AsyncGenerator[List[T], None]:
        fc = None
        if filter_dict:
            fc = cls.filter_parse(filter_dict)
        async for l in HatsudenkiClient.scan_generator(
                table_name=cls.get_collection_name(), filter_cond=fc, tick=tick, prj=prj):
            yield cls.from_raw_dict_list(l)

    @classmethod
    async def scan_list(cls: Type[T], filter_dict: dict = None, limit=20, prj: List[str] = None) -> \
            List[T]:
        l = await cls.scan_iter(filter_dict, limit, prj)
        return list(l)

    @property
    def one_cursor(self):
        h = self.__class__.get_field_class(self.hash_key_name)
        return h.to_string(self.hash_value)

    @classmethod
    def _find_index_by_query_dict(cls, query_dict: dict):
        kv = list(query_dict.keys())
        hash_key = kv[0]
        range_key = kv[1] if len(kv) > 1 else None
        op = None

        if range_key is not None:
            sp = range_key.split('__')
            if len(sp) is 2:
                range_key = sp[0]
                op = sp[1]

        use_index = cls.find_index(hash_key, range_key)
        return hash_key, range_key, op, use_index

    @classmethod
    def query_parse(cls, query_dict: dict):
        """
        クエリ
        :param query_dict:
        :return:
        """
        hash_key, range_key, op, use_index = cls._find_index_by_query_dict(query_dict)
        if use_index is None:
            raise Exception(f'invalid keys {query_dict}')

        kc = KeyConditionExpression()

        kc.equal(hash_key, query_dict[hash_key])
        if range_key is not None:
            kc.op_and()

            if op is not None:
                rv = cls.get_serialized_field_value(range_key, query_dict[range_key + '__' + op])
                kc.get_operation_by_word(op, range_key, rv, raw=True)
            else:
                rv = cls.get_serialized_field_value(range_key, query_dict[range_key])
                kc.equal(range_key, rv, raw=True)

        return kc, use_index

    @classmethod
    def filter_parse(cls, filter_dict: dict):
        fc = FilterConditionExpression()

        for k, v in filter_dict.items():
            fc.op_and()

            fc.parse_key_value(k, v)

        return fc

    def set_hash_key(self, val):
        """
        ハッシュキーの値をセットする
        このメソッドを使わずにハッシュキーを書き換えると例外が発生する

        :param val: セットする値（型はモデルに依存）
        :return: None
        """
        hk = self.get_hash_key_name()
        fc = self.__class__.get_field_class(hk)
        self.force_set_key(hk, fc.get_data(val))

    @classmethod
    async def get_or_create(cls: Type[T], hash_val: any, range_val: any = None) -> Tuple[bool, T]:
        """
        getを行い、アイテムが存在しなければ新規インスタンスを生成して返却する。
        この際、DBへの保存も同時に行われるので注意すること。

        :param hash_val: ハッシュキーの値
        :return: [0]新規作成が行われた場合は真、[1]生成された、もしくはDBから取得されたモデルインスタンス
        """
        cls._check_key(hash_val, range_val)

        ser = cls.get_serialized_key(hash_val, range_val)
        res = await cls.get_raw(ser)
        if res is None:
            r = cls()
            r.set_hash_key(hash_val)
            if range_val:
                r.set_range_key(range_val)
            await r.put()
            return True, r
        else:
            return False, res

    @classmethod
    async def delete(cls, hash_val: any, range_val: any = None):
        cls._check_key(hash_val, range_val)

        key = cls.get_serialized_key(hash_val, range_val)
        cond = cls.exist_condition()
        await HatsudenkiClient.delete_item(cls.get_collection_name(), key, cond)

    async def remove(self):
        await self.__class__.delete(self.hash_value)

    def _set_projection(self, prj: List[str]):
        """
        指定したプロジェクション情報に応じて不要な要素の削除
        :param prj: 対象キー配列
        """
        if prj is None:
            return

        # _v は必須
        prj.append('_v')
        [self.__delattr__(key) for key in self._attributes.keys() if key not in prj]

    @classmethod
    def get_primary_key_names(cls):
        return [cls.get_hash_key_name()]

    @classmethod
    def direct_update(cls):
        from hatsudenki.packages.direct.update import DirectUpdate
        d = DirectUpdate(cls)
        return d

    def begin_update(self):
        d = self.__class__.direct_update()
        d.set_key_raw(self.serialized_key)
        self.exist_condition(d.cond)
        return d

    @classmethod
    def _check_key(cls, hash_val: any, range_val: any):
        if range_val is not None:
            raise Exception('invalid key')

    def to_dict(self):
        return {k: self[k] for k, _ in self._attributes.items()}

    async def put_or_update(self):
        if self._is_new:
            await self.put()
        else:
            await self.update()

    @classmethod
    def get_serialized_field_value(cls, field_name: str, value: any):
        fc = cls.get_field_class(field_name)
        return fc.serialize(value)

    @property
    def is_modified_record(self):
        return len(self._update_keys) > 0

    @property
    def is_created_record(self):
        return self._is_new



import os
from datetime import datetime
from pathlib import Path

from hatsudenki.commands.base import BaseCommand
from hatsudenki.packages.command.files import yaml_write


class Command(BaseCommand):
    async def handle(self, *args, **options):
        c = Path(os.getcwd())

        dsl_path = c / 'dsl'

        if dsl_path.exists():
            print(f'{dsl_path} is already exists. skipped...')
            return

        dsl_path.mkdir(parents=True, exist_ok=True)

        master_path = dsl_path / 'master'
        enum_path = dsl_path / 'enum'
        dynamo_path = dsl_path / 'dynamo'
        define_path = dsl_path / 'define'

        master_path.mkdir(parents=True, exist_ok=True)
        enum_path.mkdir(parents=True, exist_ok=True)
        dynamo_path.mkdir(parents=True, exist_ok=True)
        define_path.mkdir(parents=True, exist_ok=True)

        all_tag = {
            'name': 'ALL',
            'comment': 'all output (for development)',
            'release_at': 'nothing',
            'define_at': datetime.now().isoformat()
        }
        yaml_write((define_path / 'master_tag.yml'), [all_tag])

        sc = {
            "attributes": {
                "user_id": {
                    "hash": True,
                    "type": "uuid",
                    "comment": "unique user id (UUID4)"
                },
                "kind": {
                    "range": True,
                    "type": "string",
                    "comment": "This sentence will not be used for processing"
                }
            },
            "comment": "Test table schema root definition"
        }
        yaml_write((dynamo_path / 'example' / 'root.yml'), sc)

        sc = {
            "attributes": {
                "one_number_value": {
                    "type": "number",
                    "comment": "simple number value"
                },
                "one_string_value": {
                    "type": "string",
                    "comment": "simple string value"
                }
            },
            "comment": "child solo table sample."
        }
        yaml_write((dynamo_path / 'example' / 'one.yml'), sc)

        sc = {
            "attributes": {
                "child_range_value": {
                    "type": "number",
                    "range": True,
                    "comment": "child table's range key"
                },
                "ref_master": {
                    "type": "master_one",
                    "to": "example",
                    "comment": "reference to master table(to 'master_example' table)"
                },
                "enum_value": {
                    "type": "enum",
                    "to": "example",
                    "comment": "referece to enum (to 'example' enum)"
                }
            },
            "comment": "sample child multi table."
        }
        yaml_write((dynamo_path / 'example' / 'child.yml'), sc)

        sc = {
            "name": "example",
            "origin": 0,
            "values": [
                {
                    "name": "none",
                    "ref_name": "nothing"
                },
                {
                    "name": "sample_one",
                    "ref_name": "sample_one"
                },
                {
                    "name": "sample_two",
                    "ref_name": "sample_two"
                }
            ]
        }
        yaml_write((enum_path / 'example.yml'), sc)

        sc = {
            "name": "sample",
            "excel": "sample_master_input_excel",
            "column": {
                "id": {
                    "type": "string",
                    "name": "unique_id",
                    "hash": True
                },
                "enum_sample": {
                    "name": "reference to example enum",
                    "type": "enum",
                    "to": "example"
                },
                "relation_sample": {
                    "name": "reference to master relation example",
                    "type": "master",
                    "to": "relation_example"
                }
            }
        }
        yaml_write((master_path / 'example.yml'), sc)

        sc = {
            "name": "relation_example",
            "excel": "sample_master_input_excel",
            "column": {
                "id": {
                    "type": "string",
                    "name": "unique_id",
                    "hash": True
                },
                "number_value": {
                    "type": "number",
                    "name": "number value"
                }
            }
        }
        yaml_write((master_path / 'relation_example.yml'), sc)



from hatsudenki.define.config import TAG_SEPARATOR
from hatsudenki.packages.command.hatsudenki.field import HatsudenkiFieldBase, field
from hatsudenki.packages.command.stdout.output import snake_to_camel, IndentString


@field('keys', 'str')
class MasterKeysColumn(HatsudenkiFieldBase):
    PythonStr = 'field.MasterKeysField'
    # keysは個別のクラス生成が必要
    HasGenClass = True

    def get_target_table_class_name(self, idx: int):
        t = self.data['targets'][idx]
        if self.is_master(t):
            return 'masters.' + snake_to_camel(t)
        if self.is_enum(t):
            return 'def_enum.' + snake_to_camel(t)
        return snake_to_camel(t)

    def is_master(self, key: str):
        return key.startswith('master_')

    def is_enum(self, key: str):
        return key.startswith('enum_')

    @property
    def is_alias_key(self):
        return self.is_range and not self.parent.is_alone

    @property
    def targets(self):
        return self.data['targets']

    def def_python_field_class(self, extra_opt: dict = None):
        targets = self.data['targets']
        length = len(targets)
        return f'{snake_to_camel(self.name)}Field(name="{self.name}", split_num={length})'

    @property
    def python_value_type_str(self):
        if self.is_alias_key:
            return f'{self.parent.class_name}.Meta.{snake_to_camel(self.name)}Field.Value'
        else:
            return f'{self.parent.class_name}.Field.{snake_to_camel(self.name)}Field.Value'

    @property
    def python_init_type_str(self):
        return 'list'

    @property
    def resolver_string(self):
        if not self.is_alias_key:
            return None
        if self.is_range:
            # 子テーブル且つレンジキーの場合はエイリアスキーとなる
            l = len(self.parent.tag_name) + len(TAG_SEPARATOR)
            parent = self.parent.parent_table
            parent_rk_name = parent.range_key.name

            body = IndentString()

            # getter
            body.add('@property')
            body.indent(f'def {self.name}(self) -> {self.python_value_type_str}:')
            body.indent(f'if self.{parent_rk_name}:')
            body.add(f'd = self.Meta.key_alias_type.get_data(self.{parent_rk_name}[{l}:])')
            body.outdent()
            body.indent('else:')
            body.add('d = self.Meta.key_alias_type.get_data(None)')
            body.outdent()
            body.add('d.parent = self')
            body.add('return d')
            body.outdent()

            body.blank_line()

            # setter
            body.add(f'@{self.name}.setter')
            body.indent(f'def {self.name}(self, val):')
            body.indent('if val is None:')
            body.add('return')
            body.outdent()
            body.indent('if val.is_empty():')
            body.add('return')
            body.outdent()
            body.add(
                f'self.{parent_rk_name} = f"{self.parent.tag_name}{TAG_SEPARATOR}{{val}}"')
            body.outdent()

            return body
        return None

    @property
    def class_string(self):
        u = IndentString(f'class {snake_to_camel(self.name)}Field(field.MasterKeysField):')

        u.indent('class Value(field.MasterKeysField.Value):')

        acce = IndentString()
        excep = IndentString()
        opt_strings = []

        con_body = IndentString()
        for idx, t in enumerate(self.targets):
            acce.add('@property')
            acce.indent(f'def t{idx}(self):')
            c = self.get_target_table_class_name(idx)

            if self.is_enum(t):
                acce.add(f'return {c}(int(self.label_list[{idx}])) if self.label_list[{idx}] else None')
            else:
                acce.add(f'return self.label_list[{idx}]')
            acce.outdent('')

            if self.is_master(t):
                acce.indent(f'def resolved_t{idx}(self):')
                acce.add(f'return {c}.get_by_cursor(self.t{idx})')
                acce.outdent('')
            elif self.is_enum(t):
                # enumにリゾルバは存在しない
                pass
            else:
                acce.indent(f'async def resolved_t{idx}(self):')
                acce.add(f'return await {c}.query_by_cursor(self.t{idx})')
                acce.outdent('')

            excep.indent(f'if not isinstance(t{idx}, {c}):')
            excep.add(f'raise Exception(f"not [{c}] given {{type(t{idx}).__name__}}")')
            excep.outdent()
            opt_strings.append(f't{idx}: {c}')

            if self.is_enum(t):
                con_body.add(f'self.label_list[{idx}] = str(t{idx}.value)')
            else:
                con_body.add(f'self.label_list[{idx}] = t{idx}.one_cursor')

        if not self.parent.is_alone:
            con_body.add('self.update_key()')

        con_head = IndentString(f'def connect(self, {", ".join(opt_strings)}):')
        con_head.add(excep)
        con_head.add(con_body)

        u.add(acce, con_head)

        return u



from hatsudenki.define.config import TAG_SEPARATOR
from hatsudenki.packages.command.hatsudenki.field import field, HatsudenkiFieldBase
from hatsudenki.packages.command.stdout.output import snake_to_camel, IndentString


@field('dynamo_one', 'str')
class ReferenceDynamoOneField(HatsudenkiFieldBase):
    PythonStr = 'field.ReferenceDynamoOneField'

    def _parse_opt(self):
        o = super()._parse_opt()
        o.append(f"to='{self.target_table_name}'")
        return o

    @property
    def field_type(self):
        return f'field.ReferenceDynamoOneField'

    @property
    def python_value_type_str(self):
        return f'{self.field_type}.Value[{self.target_table_class_name}]'

    def def_python_field_class(self, extra_opt: dict = None):
        return f'{self.field_type}({self.get_init_option(extra_opt)})'

    @property
    def target_table_name(self) -> str:
        return self.data['to']

    @property
    def target_table_class_name(self):
        return snake_to_camel(self.target_table_name)

    @property
    def resolver_string(self):
        if not self.is_alias_key:
            return None
        if self.is_range:
            # 子テーブル且つレンジキーの場合はエイリアスキーとなる
            l = len(self.parent.tag_name) + len(TAG_SEPARATOR)
            parent = self.parent.parent_table
            parent_rk_name = parent.range_key.name

            body = IndentString()

            # getter
            body.add('@property')
            body.indent(f'def {self.name}(self) -> {self.python_value_type_str}:')
            body.indent(f'if self.{parent_rk_name}:')
            body.add(f'd = self.Meta.key_alias_type.get_data(self.{parent_rk_name}[{l}:])')
            body.outdent()
            body.indent('else:')
            body.add('d = self.Meta.key_alias_type.get_data(None)')
            body.outdent()
            body.add('d.parent = self')
            body.add('return d')
            body.outdent()

            body.blank_line()

            # setter
            body.add(f'@{self.name}.setter')
            body.indent(f'def {self.name}(self, val):')
            body.indent('if val is None:')
            body.add('return')
            body.outdent()
            body.indent('if val.is_empty():')
            body.add('return')
            body.outdent()
            body.add(
                f'self.{parent_rk_name} = f"{self.parent.tag_name}{TAG_SEPARATOR}{{val}}"')
            body.outdent()

            return body
        return None


@field('dynamo_many', 'str')
class ReferenceDynamoManyField(ReferenceDynamoOneField):
    PythonStr = 'field.ReferenceDynamoManyField'

    @property
    def field_type(self):
        return f'field.ReferenceDynamoManyField'


@field('master_one', 'str')
class ReferenceMasterOneField(HatsudenkiFieldBase):
    PythonStr = 'field.ReferenceMasterOneField'

    def gen_init_str(self, direct_assign=True):
        if direct_assign:
            if self.is_alias_key:
                return f'self.{self.name} = self.Meta.key_alias_type.get_data({self.name}, self)'
            elif self.is_hash or self.is_range:
                return f'self.{self.name} = ft.{self.name}.get_data({self.name}, self)'
        return f'self.{self.name} = ft.{self.name}.get_data_from_dict(kwargs, self)'

    def _parse_opt(self):
        o = super()._parse_opt()
        o.append(f"to=masters.{self.target_table_class_name}")
        return o

    @property
    def field_type(self):
        return f'field.ReferenceMasterOneField'

    @property
    def python_init_type_str(self):
        return 'masters.' + self.target_table_class_name

    @property
    def python_value_type_str(self):
        return f'{self.field_type}.Value'

    def def_python_field_class(self, extra_opt: dict = None):
        return f'{self.field_type}({self.get_init_option(extra_opt)})'

    @property
    def target_table_name(self) -> str:
        return self.data['to']

    @property
    def target_table_class_name(self):
        return 'Master' + snake_to_camel(self.target_table_name)

    @property
    def resolver_string(self):
        if not self.is_alias_key:
            return None
        if self.is_range:
            # 子テーブル且つレンジキーの場合はエイリアスキーとなる
            l = len(self.parent.tag_name) + len(TAG_SEPARATOR)
            parent = self.parent.parent_table
            parent_rk_name = parent.range_key.name

            body = IndentString()

            # getter
            body.add('@property')
            body.indent(f'def {self.name}(self):')
            body.indent(f'if self.{parent_rk_name}:')
            body.add(f'd = self.Meta.key_alias_type.get_data(self.{parent_rk_name}[{l}:])')
            body.outdent()
            body.indent('else:')
            body.add('d = self.Meta.key_alias_type.get_data(None)')
            body.outdent()
            body.add('d.parent = self')
            body.add('return d')
            body.outdent()

            body.blank_line()

            # setter
            body.add(f'@{self.name}.setter')
            body.indent(f'def {self.name}(self, val):')
            body.indent('if val is None:')
            body.add('return')
            body.outdent()
            body.indent('if val.is_empty():')
            body.add('return')
            body.outdent()
            body.add(
                f'self.{parent_rk_name} = f"{self.parent.tag_name}{TAG_SEPARATOR}{{val}}"')
            body.outdent()

            return body
        return None


@field('master_many', 'str')
class ReferenceMasterManyField(ReferenceMasterOneField):
    PythonStr = 'field.ReferenceMasterManyField'

    @property
    def field_type(self):
        return f'field.ReferenceMasterManyField'



from pathlib import Path
from typing import List, Dict

from hatsudenki.packages.command.files import yaml_load
from hatsudenki.packages.command.master.column import ColumnChose
from hatsudenki.packages.command.master.table import MasterTable
from hatsudenki.packages.command.stdout.output import ToolOutput
from hatsudenki.packages.command.util.base_info import BaseInfo


class MasterDataYaml(BaseInfo):
    def __init__(self, base_path: Path, full_path: Path, table: MasterTable, loader, *args, **kwargs):
        from hatsudenki.packages.command.master.importer.loader import MasterDataYamlLoader
        super().__init__(base_path, full_path, *args, **kwargs)
        self.data: List[Dict[str, any]] = yaml_load(full_path)
        self.table = table
        self.loader: MasterDataYamlLoader = loader

    @property
    def table_name(self):
        return self.table.table_name

    def iter(self):
        ToolOutput.anchor()
        for data in self.data:
            ret = {}
            for key, col in data.items():
                ToolOutput.anchor()
                col_scheme = self.table.columns.get(key)
                if col_scheme is None:
                    ToolOutput.print(f'{key} が見つかりません "{self.full_path}:0"\n"{self.table.full_path}:0"')
                    ToolOutput.pop()
                    continue

                if isinstance(col_scheme, ColumnChose):
                    # Choseはselectorの値が無いとExcel上の値が確定しない
                    ret[col_scheme.excel_header_name] = col_scheme.resolve_value(data[col_scheme.selector], col,
                                                                                 is_raw=True, ret_ref_name=True)
                else:
                    ret[col_scheme.excel_header_name] = col_scheme.reverse_value(col)
                ToolOutput.pop()
            yield ret
        ToolOutput.pop()



import bisect
from typing import Type, Iterator, List, Iterable

from hatsudenki.packages.cache.base.define import CURSOR_SEPARATOR
from hatsudenki.packages.cache.base.memory.solo import CacheBaseTableSolo, T


class CacheBaseTableMulti(CacheBaseTableSolo):
    _sorted_table = {}

    @classmethod
    def _get(cls: Type[T], hash_key, range_key=None) -> T:
        if range_key is None:
            raise Exception('invalid key.')
        l = cls._sorted_table[hash_key]
        l_idx = bisect.bisect_left(l, range_key)
        if l_idx != len(l) and l[l_idx] == range_key:
            return cls._mapped[hash_key][l_idx]
        raise Exception(f'data not found. {cls.Meta.table_name} {hash_key} {range_key}')

    @classmethod
    def _find_left(cls: Type[T], hash_key, range_key=None) -> T:
        """
        たとえば[0,100,200,300]のリストに対して下記のように探索する
        50: 0
        100: 1
        150: 1
        200: 2

        :param hash_key:
        :param range_key:
        :return:
        """
        if range_key is None:
            raise Exception('invalid key.')
        l = cls._sorted_table[hash_key]
        l_idx = bisect.bisect_left(l, range_key)
        list_length = len(l)
        if l_idx >= list_length:
            l_idx = list_length - 1
        if range_key < l[l_idx]:
            l_idx -= 1
        if l_idx < 0:
            return cls._mapped[hash_key][0]
            # raise Exception(f'data not found. {cls.Meta.table_name} {hash_key} {range_key} {l_idx}')
        return cls._mapped[hash_key][l_idx]

    @classmethod
    def _find_right(cls: Type[T], hash_key, range_key=None) -> T:
        """
        たとえば[100,200,300]のリストに対して下記のように探索する
        50: 0
        100: 1
        150: 1
        200: 2

        :param hash_key:
        :param range_key:
        :return:
        """
        if range_key is None:
            raise Exception('invalid key.')
        l = cls._sorted_table[hash_key]
        r_idx = bisect.bisect_right(l, range_key)
        list_length = len(l)
        if r_idx >= list_length:
            r_idx = list_length - 1
        return cls._mapped[hash_key][r_idx]

    @classmethod
    def _find(cls: Type[T], hash_key) -> List[T]:
        """
        HASHキーのみを指定してリスト
        :param hash_key: ハッシュキー
        :return: レコードの配列
        """
        l = cls._mapped[hash_key]
        return l

    @classmethod
    def _find_lte(cls: Type[T], hash_key: str, range_key) -> List[T]:
        """
        RANGEキー条件指定検索 - 以下

        :param hash_key: ハッシュキー
        :param range_key: 対象値
        :return: レコードの配列
        """
        f = cls._sorted_table[hash_key]
        i = bisect.bisect_right(f, range_key)
        return cls._mapped[hash_key][0:i]

    @classmethod
    def _find_lt(cls: Type[T], hash_key: str, range_key) -> List[T]:
        """
        RANGEキー条件指定検索 - 未満

        :param hash_key: ハッシュキー
        :param range_key: 対象値
        :return: レコードの配列
        """
        f = cls._sorted_table[hash_key]
        i = bisect.bisect_left(f, range_key)
        return cls._mapped[hash_key][0:i]

    @classmethod
    def _find_gte(cls: Type[T], hash_key: str, range_key) -> List[T]:
        """
        RANGEキー条件指定検索 - 以上

        :param hash_key: ハッシュキー
        :param range_key: 対象値
        :return: レコードの配列
        """
        f = cls._sorted_table[hash_key]
        i = bisect.bisect_left(f, range_key)
        return cls._mapped[hash_key][i:]

    @classmethod
    def _find_gt(cls: Type[T], hash_key: str, range_key) -> List[T]:
        """
        RANGEキー条件指定検索 - より小さい

        :param hash_key: ハッシュキー
        :param range_key: 対象値
        :return: レコードの配列
        """
        f = cls._sorted_table[hash_key]
        i = bisect.bisect_right(f, range_key)
        return cls._mapped[hash_key][i:]

    @classmethod
    def _find_between(cls: Type[T], hash_key: str, left_range, right_range) -> List[T]:
        """
        RANGEキー条件指定検索 - 範囲

        :param hash_key: ハッシュキー
        :param left_range: 左辺値
        :param right_range: 右辺値
        :return: レコードの配列
        """
        f = cls._sorted_table[hash_key]
        l = bisect.bisect_left(f, left_range)
        r = bisect.bisect_right(f, right_range)
        return cls._mapped[hash_key][l:r]

    @classmethod
    def _get_by_cursor(cls: Type[T], cursor: str) -> T:
        sp = cursor.split(CURSOR_SEPARATOR, 2)
        return cls._get(*sp)

    @classmethod
    def _find_by_cursor(cls: Type[T], cursor: str) -> List[T]:
        """
        集合カーソルを指定してリストを取得

        :param cursor: str
        :return: レコードの配列
        """
        return cls._find(cursor)

    @classmethod
    def _iter(cls: Type[T]) -> Iterator[T]:
        for _, l in cls._mapped.items():
            for d in l:
                yield d

    @classmethod
    def get_range_key_name(cls):
        """
        RANGEキーの名前を取得する

        :return: str
        """
        return cls.Meta.primary_index.range_key

    @property
    def one_cursor(self):
        return str(self.hash_value) + CURSOR_SEPARATOR + str(self.range_value)

    @property
    def many_cursor(self):
        """
        集合を表す文字列

        :return: str
        """
        return str(self.hash_value)

    @property
    def range_value(self):
        """
        RANGEキーの値を取得

        :return: RANGEキーに設定されている値
        """
        return getattr(self, self.__class__.get_range_key_name())

    @classmethod
    def _set_records(cls, recs: Iterable[dict]):
        hk = cls.get_hash_key_name()
        rk = cls.get_range_key_name()
        d: dict = {}
        s: dict = {}

        for rec in recs:
            hk_val = rec[hk]
            rk_val = rec[rk]
            n = d.get(hk_val, [])
            idx = bisect.bisect_right(n, rk_val)
            n.insert(idx, rk_val)
            d[hk_val] = n

            ss = s.get(hk_val, [])
            ss.insert(idx, cls(**rec))
            s[hk_val] = ss

        cls._mapped = s
        cls._sorted_table = d

    @classmethod
    def _append(cls: Type[T], rec: T):
        hk = cls.get_hash_key()
        rk = cls.get_range_key()
        hk_val = rec[hk]
        rk_val = rec[rk]
        n = cls._sorted_table.get(hk_val, [])
        idx = bisect.bisect_left(n, rk_val)

        n.insert(idx, rk_val)
        cls._sorted_table[hk_val] = n

        ss = cls._mapped.get(hk_val, [])
        ss.insert(idx, cls(**rec))
        cls._mapped[hk_val] = ss



from hatsudenki.packages.command.master.column import MasterColumn
from hatsudenki.packages.command.master.loader import MasterTableLoader
from hatsudenki.packages.command.master.table import MasterTable
from hatsudenki.packages.command.renderer.base import RenderUnit, FileRenderer
from hatsudenki.packages.command.stdout.output import IndentString, quote


class MasterDataStoreRenderUnit(RenderUnit[MasterTable]):
    def _model(self):
        u = IndentString()
        m = IndentString()
        f = IndentString()

        if self.data.range_key is not None:
            u.indent(f'class {self.data.class_name}(MasterModelMulti):')
        else:
            u.indent(f'class {self.data.class_name}(MasterModelSolo):')

        m.indent('class Meta:')
        m.add(f"table_name = '{self.data.table_name}'")
        m.add(
            f"primary_index = PrimaryIndex({', '.join([f'{quote(l.python_name)}' for l in self.data.cursor_labels])})")

        f.indent('class Field:')

        i = IndentString()
        i.indent('def __init__(self, **kwargs):')
        i.add('super().__init__(**kwargs)')
        i.add('fields = self.__class__.Field')

        resolver = IndentString()
        no_serial = True
        for k, c in self.data.columns.items():
            f.add(c.python_define_name)
            i.add(c.python_init_name)

            if c.type_name == 'master':
                # master
                resolver.add(c.resolver_name)
                resolver.blank_line()
            elif c.type_name == 'chose':
                # select/choseの解決
                resolver.add(c.resolver_name(self.data.columns[c.selector]))
                resolver.blank_line()
            elif c.is_serial_key:
                # ログ用シリアル
                resolver.add(c.serial_resolver_name)
                no_serial = False
        # シリアルがないテーブルではone_cursorを返却する
        if no_serial:
            resolver.add(MasterColumn.serial_resolver_default())

        m.blank_line()
        u.add(m)
        f.blank_line()
        u.add(f)
        u.add(i)

        if resolver.line_num > 0:
            u.blank_line()
            u.add(resolver)

        u.blank_line()
        return u

    def render(self):
        u = self._model()
        return u


class MasterModelRenderer(FileRenderer):

    def __init__(self, master_loader: MasterTableLoader):
        super().__init__()
        self.master_loader = master_loader

    def render_header(self) -> IndentString:
        u = IndentString()
        u.add('from hatsudenki.packages.cache.base import column')
        u.add('from hatsudenki.packages.cache.base.index import PrimaryIndex')
        u.add('from hatsudenki.packages.master.model import MasterModelSolo, MasterModelMulti')
        u.blank_line(2)
        return u

    def render_units(self):
        u = IndentString()
        [u.add(MasterDataStoreRenderUnit(table).render()) for k, table in self.master_loader.iter()]
        return u



from typing import Type, TypeVar, Tuple, List

from hatsudenki.define.config import TAG_SEPARATOR, CURSOR_SEPARATOR
from hatsudenki.packages.field.base import BaseHatsudenkiField
from hatsudenki.packages.table.define import TableType
from hatsudenki.packages.table.multi import MultiHatsudenkiTable

T = TypeVar("T")


class ChildMultiHatsudenkiTable(MultiHatsudenkiTable):
    class Meta(MultiHatsudenkiTable.Meta):
        key_alias_name = ''
        key_alias_type: BaseHatsudenkiField = None
        tag_name = ''

    @classmethod
    def get_table_type(cls):
        return TableType.ChildTable

    @classmethod
    def get_alias_key(cls):
        return cls.Meta.key_alias_name

    @property
    def alias_value(self):
        return getattr(self, self.__class__.get_alias_key())

    def set_alias_key(self, val):
        a = self.resolve_alias(val)
        self.set_range_key(a)

    @classmethod
    def get_alias_key_type(cls):
        return cls.Meta.key_alias_type

    @classmethod
    def serialize_alias_key(cls, val):
        return cls.get_alias_key_type().serialize(val)

    @classmethod
    def get_tag_name(cls):
        return cls.Meta.tag_name

    @classmethod
    def generate_range_key_str(cls, val=None):
        f = cls.get_alias_key_type()
        return f'{cls.get_tag_name()}{TAG_SEPARATOR}{f.to_string(val)}'

    @classmethod
    def resolve_alias(cls, val):
        if val is None:
            raise Exception('invalid key.')
        r = cls.generate_range_key_str(val)
        return r

    @classmethod
    def query_parse(cls, query_dict: dict):
        # childテーブルはkindを自動設定する

        # 使用するインデックスを判定
        ak = cls.get_alias_key()
        rk = cls.get_range_key_name()
        hash_key, range_key, op, use_index = cls._find_index_by_query_dict(query_dict)
        if range_key == ak:
            if op is None:
                # エイリアスキーが明示的に指定されている
                query_dict[rk] = cls.resolve_alias(query_dict[ak])
                # エイリアスキーは実在しないので消す
                del query_dict[ak]
            elif op == 'beginsWith':
                # エイリアスキーが指定され、且つオペレータが指定されている
                k = ak + '__' + op
                v = query_dict[k]
                # エイリアスキーは実在しないので消す
                query_dict[rk + '__' + op] = cls.resolve_alias(v)
                del query_dict[ak + '__' + op]
            else:
                # ごめんね、まだできてないんだ
                raise Exception(f'alias_key operation [{op}] is not supported yet.')

        elif range_key is None and hash_key == cls.get_hash_key_name():
            # レンジキーが指定されていない且つ、ハッシュキーがプライマリーのもの（GSIでない）場合は
            # 暗黙的に本来のレンジキーをTAG_NAMEの前方一致検索とする
            query_dict[cls.get_range_key_name() + '__beginsWith'] = f'{cls.get_tag_name()}{TAG_SEPARATOR}'

        return super().query_parse(query_dict)

    @property
    def many_cursor(self):
        h = self.get_hash_field_class()
        return h.to_string(self.hash_value)

    @property
    def one_cursor(self):
        h = self.get_hash_field_class()
        t = self.get_alias_key_type()
        return f'{h.to_string(self.hash_value)}{CURSOR_SEPARATOR}{t.to_string(self.alias_value)}'

    @classmethod
    async def get_or_create(cls: Type[T], hash_val: any, alias_val: any = None) -> Tuple[bool, T]:
        r = cls.resolve_alias(alias_val)
        return await super().get_or_create(hash_val, r)

    @classmethod
    async def get(cls: Type[T], hash_val: any, alias_val: any = None, prj_exp: List[str] = None) -> T:
        range_val = cls.resolve_alias(alias_val)
        return await super().get(hash_val, range_val, prj_exp)

    @classmethod
    async def delete(cls, hash_val: any, alias_val: any = None):
        range_val = cls.resolve_alias(alias_val)
        return await super().delete(hash_val, range_val)

    async def remove(self):
        await self.__class__.delete(self.hash_value, self.alias_value)

    async def reget(self):
        c = await self.__class__.get(self.hash_value, self.alias_value)
        self.copy(c)



from pathlib import Path
from typing import List, Dict

from hatsudenki.packages.command.files import yaml_load
from hatsudenki.packages.command.master.table import MasterTable
from hatsudenki.packages.command.util.base_info import BaseInfo


class MasterExportYaml(BaseInfo):
    def __init__(self, base_path: Path, full_path: Path, table: MasterTable, loader, *args, **kwargs):
        from hatsudenki.packages.command.master.exporter.pack.loader import MasterExportYamlLoader
        super().__init__(base_path, full_path, *args, **kwargs)
        self.data: List[Dict[str, any]] = yaml_load(full_path)
        self.table = table
        self.loader: MasterExportYamlLoader = loader
        self._cache = None
        self._list_cache = None

    @property
    def table_name(self):
        return self.table.table_name

    def iter(self):
        return (d for d in self.data)

    def find(self, hash_val: any, range_val: any = None) -> Dict[str, any]:
        hash_key = self.table.hash_key
        range_key = self.table.range_key

        if self._cache is None:
            self._cache = {}
            for rec in self.data:
                if range_key is None:
                    cache_key = (rec[hash_key.column_name],)
                else:
                    cache_key = (rec[hash_key.column_name], rec[range_key.column_name])

                self._cache[cache_key] = rec

        if range_key is None:
            find_key = (hash_val,)
        else:
            find_key = (hash_val, range_val)

        return self._cache[find_key]

    def find_list(self, hash_val: any) -> List[Dict[str, any]]:
        hash_key = self.table.hash_key

        if self._list_cache is None:
            self._list_cache = {}
            for rec in self.data:
                list_cache_key = rec[hash_key.column_name]
                self._list_cache.setdefault(list_cache_key, [])
                self._list_cache[list_cache_key].append(rec)

        return self._list_cache[hash_val]



from typing import Generic, TypeVar

from hatsudenki.packages.expression.update import UpdateExpression
from hatsudenki.packages.field.util import primal_deserializer, primal_serializer

T = TypeVar('T')


class BaseHatsudenkiField(Generic[T]):
    """
    DynamoDBのアイテムフィールドクラス
    """
    PythonType = None
    TypeStr = ''
    ValueType = None
    DefaultValue = None
    TypeName = ''
    IsScalar = True

    def __init__(self, *, name=None, default=None, ttl=False):
        """
        init

        :param name: 名前
        :param default: デフォルト値
        :param ttl: ttlとして参照するか
        """
        self.is_ttl = ttl
        self._default_value = default
        self.name = name

    @classmethod
    def is_empty(cls, value):
        """
        値が空か

        :param value: 対象値
        :return: bool
        """
        return value is None

    @property
    def default_value(self):
        """
        デフォルト値

        :return: value
        """
        return self._default_value if self._default_value is not None else self.__class__.DefaultValue

    @property
    def attribute_define(self):
        """
        定義用スキーマ

        :return: dict
        """
        return {
            'AttributeName': self.name,
            'AttributeType': self.TypeStr
        }

    def key_schema(self, type_str: str):
        """
        キー定義用スキーマ

        :param type_str: 'HASH'または'RANGE'
        :return: dict
        """
        return {
            'AttributeName': self.name,
            'KeyType': type_str
        }

    def deserialize(self, value, table=None):
        """
        Pythonが解釈できる値にデシリアライズする

        :param value:
        :param table:
        :return:
        """
        if value is None:
            return self._default_value

        return primal_deserializer(value)

    def serialize(self, value: T, table=None):
        """
        DynamoDBが解釈できる値にシリアライズする

        :param value: 対象の値
        :return: DynamoDBに準拠した値情報連想配列
        """
        if self.is_empty(value):
            return None

        return primal_serializer(value)

    @classmethod
    def set_expression(cls, name: str, value, update: UpdateExpression):
        update.set(name, cls.serialize(value), raw=True)

    @classmethod
    def add_expression(cls, name: str, new_value, old_value, update: UpdateExpression):
        update.set(name, cls.serialize(new_value), raw=True)

    def __str__(self):
        return self.name

    def get_data(self, val, table=None) -> T:
        if val is None:
            return self.default_value

        if type(val) is self.PythonType:
            return val
        return self.PythonType(val)

    def get_data_from_dict(self, v: dict, table=None) -> T:
        return self.get_data(v.get(self.name, None), table)

    def dump_text(self, val):
        return f'{self.name}: {val.__repr__()}'

    def to_string(self, val):
        return str(val)

    def build_update_expression(self, upd: UpdateExpression, now_value):
        upd.set(self.name, self.serialize(now_value), raw=True)



from hatsudenki.packages.expression.update import UpdateExpression
from hatsudenki.packages.field.base import BaseHatsudenkiField


class StringSetField(BaseHatsudenkiField[set]):
    """
    文字列セットフィールドクラス。
    """

    TypeStr = 'SS'
    ValueType = str
    TypeName = 'StringSet'

    def is_empty(cls, value):
        return value is None or len(value) is 0

    @classmethod
    def add_expression(cls, name: str, new_value, old_value, update: UpdateExpression):
        update.increment(name, cls.serialize(new_value - old_value), raw=True)


class NumberSetField(BaseHatsudenkiField[set]):
    """
    数値セットフィールドクラス。重複した値は格納できない。
    """
    PythonType = set

    TypeStr = 'NS'
    ValueType = int
    TypeName = 'NumberSet'

    @classmethod
    def add_expression(cls, name: str, new_value, old_value, update: UpdateExpression):
        update.increment(name, cls.serialize(new_value - old_value), raw=True)


class BinarySetField(BaseHatsudenkiField[set]):
    """
    バイナリセットフィールドクラス。重複した値は格納できない。
    """
    PythonType = set

    TypeStr = 'BS'
    ValueType = bytes
    TypeName = 'BinarySet'

    @classmethod
    def is_empty(cls, value):
        return value is None or len(value) is 0

    @classmethod
    def add_expression(cls, name: str, new_value, old_value, update: UpdateExpression):
        update.increment(name, cls.serialize(new_value - old_value), raw=True)



import logging
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import TypeVar, Type, Dict, List

import dill
import yaml

from hatsudenki.packages.cache.base.memory.static.multi import StaticCacheBaseTableMulti
from hatsudenki.packages.cache.base.memory.static.solo import StaticCacheBaseTableSolo

_logger = logging.getLogger(__name__)

T = TypeVar('T')


class MasterModelSolo(StaticCacheBaseTableSolo):
    """
    HASHキーのみを持つマスターテーブル
    """

    _cache_dict: Dict[str, Dict[any, List[T]]] = {}
    _dump_base_path: Path = None

    class Field:
        pass

    @property
    def is_nothing(self):
        """
        [なし]データか？

        :return: boolean
        """
        return self.hash_value == 'なし'

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        from hatsudenki.packages.master.manager.table import MasterTableManager
        MasterTableManager.register(cls)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @classmethod
    def _get_by_cursor(cls: Type[T], cursor: str) -> T:
        hash_field = cls.get_hash_field_class()
        return super()._get_by_cursor(cursor)

    @classmethod
    def _load_from_yaml(cls, base_path: Path, cache_base_path: Path, force=False):
        """
        ロード処理。キャッシュが存在する場合はそちらを参照し、|
        そうでない場合はYAMLからロード後にキャッシュを生成する

        :param base_path: 検索ディレクトリパス
        :param cache_base_path: キャッシュディレクトリパス
        :param force: キャッシュを無視して強制的にYAMLを読み込む
        :return: None
        """
        cls._mapped = {}
        cls._cache_dict = {}

        yml_path = base_path / (cls.Meta.table_name + '.yml')
        cache_path = cache_base_path / (cls.Meta.table_name + '_mapped.pkl')
        _logger.info(f'load from yaml {yml_path} {cache_path}')

        # forceフラグがONの場合はキャッシュを見ない
        if not force:
            if cls._load_from_dill(cache_path):
                # キャッシュからロードに成功したのでおわり
                _logger.info(f'CACHE HIT! {cache_path} を読み込み')
                return

        # YAMLからロードし、キャッシュをダンプする
        try:
            with yml_path.open(encoding='utf-8') as file:
                ret = yaml.safe_load(file)
            _logger.info(f'{yml_path}を読み込み')
        except:
            _logger.warning(f'{yml_path}の読み込みに失敗')
            ret = []

        cls._set_records(ret)
        cls._dump_dill(cache_path)

    @classmethod
    def _load_from_dill(cls, cache_path: Path):
        """
        dillダンプからデータをロード

        :param cache_path: キャッシュパス
        :return: bool
        """

        if cache_path.exists():
            _logger.info(f'キャッシュロード {cache_path}')
            with cache_path.open(mode='rb') as f:
                cls._mapped = dill.loads(f.read())

            return True
        return False

    @classmethod
    def _dump_dill(cls, out_path: Path):
        """
        dillダンプを出力

        :param out_path: 出力先パス
        :return: None
        """
        Path(out_path.parent).mkdir(exist_ok=True, parents=True)
        with out_path.open(mode='wb') as f:
            f.write(dill.dumps(cls._mapped))

    @classmethod
    def get_dictionary(cls: Type[T], key_name: str) -> Dict[str, List[T]]:
        """
        指定したキーをハッシュとした連想配列を取得

        :param key_name: ハッシュキーとなるキー名
        :return: dict
        """

        if key_name in cls._cache_dict:
            return cls._cache_dict[key_name]

        r = {}
        for d in cls.iter():
            r.setdefault(getattr(d, key_name), [])
            r[getattr(d, key_name)].append(d)
        cls._cache_dict[key_name] = r
        return cls._cache_dict[key_name]

    @classmethod
    def find_one(cls, key_name: str, key):
        """
        キーを取得して一件取得
        この関数は指定された属性名をキーとした連想配列をキャッシュするため、一回目は重い。

        :param key_name: 対象キー名
        :param key: 検索対象の値
        :return: 条件に合致するアイテム。存在しない場合はKeyError例外をraiseする。
        :except: KeyError - みつからない
        """
        d = cls.get_dictionary(key_name)
        return d[key]

    @classmethod
    def _get(cls: Type[T], hash_key, range_key=None) -> T:
        hk = cls.get_hash_key_name()
        fc = getattr(cls.Field, hk)

        try:
            res = super()._get(fc.convert(hash_key))
        except KeyError:
            raise Exception(f'{cls.__name__} から {hash_key}:{range_key} が見つかりません')

        return res

    @classmethod
    def get_hash_field_class(cls):
        return getattr(cls.Field, cls.get_hash_key_name())

    @classmethod
    def is_multi(cls):
        return False


class MasterModelMulti(StaticCacheBaseTableMulti, MasterModelSolo):
    """
    HASHキーとRANGEキーを持つマスターテーブル
    """

    @dataclass
    class CacheData:
        mapped: dict
        sorted_table: dict

    @classmethod
    def _load_from_dill(cls, cache_path: Path):
        cls._mapped = {}
        cls._sorted_table = {}
        if cache_path.exists():
            _logger.info(f'キャッシュロード {cache_path}')
            with cache_path.open(mode='rb') as f:
                d: cls.CacheData = dill.loads(f.read())
                cls._mapped = d.mapped
                cls._sorted_table = d.sorted_table
            return True

        return False

    @classmethod
    def _dump_dill(cls, out_path: Path):
        d = cls.CacheData(mapped=cls._mapped, sorted_table=cls._sorted_table)
        Path(out_path.parent).mkdir(parents=True, exist_ok=True)
        with out_path.open(mode='wb') as f:
            f.write(dill.dumps(d))

    @classmethod
    def _get(cls: Type[T], hash_key, range_key=None) -> T:
        hk = cls.get_hash_key_name()
        fc = getattr(cls.Field, hk)
        rk = cls.get_range_key_name()
        rfc = getattr(cls.Field, rk)

        try:
            res = super()._get(fc.convert(hash_key), rfc.convert(range_key))
        except KeyError:
            raise Exception('master not found.')

        return res

    @classmethod
    def get_range_field_class(cls):
        return getattr(cls.Field, cls.get_range_key_name())

    @classmethod
    def is_multi(cls):
        return True


@dataclass
class EnumResolver:
    enum_type: Type[Enum]

    def get(self, item: str):
        return self.enum_type(int(item))


@dataclass
class MasterResolver:
    master_type: Type[MasterModelSolo]

    def get(self, item: any):
        if issubclass(self.master_type, MasterModelMulti):
            # Multiテーブルの場合はリストで返す
            return self.master_type.find(item)
        else:
            # soloテーブルの場合は単体で返す
            return self.master_type.get(item)



from __future__ import annotations

from typing import Optional, List

from attr import dataclass


@dataclass
class Schema:
    @dataclass
    class Group:
        @dataclass
        class Table:
            @dataclass
            class Column:
                name: str

                def print(self):
                    print(f"        {self.name}")

                def print_str(self):
                    return f"        - {self.name}"

            name: str
            columns: List[Column]

            def __sub__(self, other: Schema.Group.Table) -> Optional[Schema.Group.Table]:
                if other is None:
                    return self
                if self.name != other.name:
                    raise Exception('Can not sub')
                diff_column = [c for c in self.columns if c.name not in [o.name for o in other.columns]]
                if len(diff_column) == 0:
                    return None
                return Schema.Group.Table(self.name, diff_column)

            def print(self):
                print(f"    {self.name}")
                [c.print() for c in self.columns]

            def print_str(self):
                res = f"    - {self.name} \r \n" + " \r \n".join([c.print_str() for c in self.columns])
                return res


        name: str
        tables: List[Table]

        def __sub__(self, other: Schema.Group) -> Optional[Schema.Group]:
            if other is None:
                return self
            if self.name != other.name:
                raise Exception('Can not sub')
            diff_gen = (s - next((t for t in other.tables if t.name == s.name), None) for s in self.tables)
            diff_tbls = [t for t in diff_gen if t is not None]
            if len(diff_tbls) == 0:
                return None
            return Schema.Group(self.name, diff_tbls)

        def print(self):
            print(f"{self.name}")
            [t.print() for t in self.tables]

        def print_str(self):
            return f" - {self.name} \r \n" + " \r \n".join([t.print_str() for t in self.tables])

    groups: List[Group]

    def __sub__(self, other: Schema) -> Optional[Schema]:
        diff_gen = (g - next((o for o in other.groups if o.name == g.name), None) for g in self.groups)
        diff = [g for g in diff_gen if g is not None]
        if len(diff) == 0:
            return None
        return Schema(diff)

    def print(self):
        [g.print() for g in self.groups]

    def print_str(self):
        return " \r \n".join([g.print_str() for g in self.groups])



from logging import getLogger
from typing import TypeVar, Type

import dill

from hatsudenki.packages.table.index import PrimaryIndex

T = TypeVar('T')
_logger = getLogger(__name__)


class InOutCacheBaseTable(object):
    class Meta:
        table_name: str = ''
        index: PrimaryIndex = None

    def serialize(self):
        return dill.dumps(self)

    @classmethod
    def deserialize(cls: Type[T], b: any) -> T:
        return dill.loads(b)

    @classmethod
    def get_table_name(cls):
        return cls.Meta.table_name

    @classmethod
    def get_hash_key(cls):
        return cls.Meta.index.hash_key

    @classmethod
    def get_range_key(cls):
        return cls.Meta.index.range_key

    @classmethod
    def _gen_key(cls, hash_val, range_val=None):
        k = f'{cls.get_table_name()}:{str(hash_val)}'

        if range_val:
            k += '--' + str(range_val)
        return k

    def _get_key(self):
        cls = self.__class__
        hk = cls.get_hash_key()
        rk = cls.get_range_key()

        hv = getattr(self, hk)
        rv = getattr(self, rk, None) if rk else None
        return cls._gen_key(hv, rv)

    async def save(self):
        pass

    @classmethod
    async def get(cls: Type[T], hash_value, range_value=None) -> T:
        v = await cls._get(hash_value, range_value)
        if v is None:
            return None
        return cls.deserialize(v)

    @classmethod
    async def _get(cls, hash_key, range_key=None) -> bytes:
        pass



from hatsudenki.packages.cache.base.memory.dynamic.multi import DynamicCacheBaseTableMulti
from hatsudenki.packages.cache.base.memory.dynamic.solo import DynamicCacheBaseTableSolo
from hatsudenki.packages.cache.base.memory.dynamic.sync.multi import SyncDynamicCacheBaseTableMulti
from hatsudenki.packages.cache.base.memory.dynamic.sync.solo import SyncDynamicCacheBaseTableSolo


class InMemoryTableSolo(DynamicCacheBaseTableSolo):
    def __init_subclass__(cls, **kwargs):
        from hatsudenki.packages.cache.inmemory.manager import InMemoryTableManager
        InMemoryTableManager.register(cls)


class InMemoryTableMulti(DynamicCacheBaseTableMulti, DynamicCacheBaseTableSolo):
    pass


class SyncInMemoryTableSolo(SyncDynamicCacheBaseTableSolo):
    def __init_subclass__(cls, **kwargs):
        from hatsudenki.packages.cache.inmemory.manager import InMemoryTableManager
        InMemoryTableManager.register(cls)


class SyncInMemoryTableMulti(SyncDynamicCacheBaseTableMulti, DynamicCacheBaseTableSolo):
    pass



from typing import Dict

from hatsudenki.packages.expression.update import UpdateExpression
from hatsudenki.packages.marked import SerializableField, Markable


class MarkedObject(object):
    """
    更新されたアトリビュートを保持し、更新箇所だけ変更するクエリをビルドできるオブジェクト
    """
    # フィールドクラスを文字列で引けるようにするための参照キャッシュ
    __fields: Dict[str, SerializableField] = {}

    class Field:
        """
        継承先ではここにフィールドを定義する
        """
        pass

    def __init_subclass__(cls, **kwargs):
        # フィールドを文字列で参照できるようにキャッシュしておく
        cls.__fields = {}
        for k in dir(cls.Field):
            if k.startswith('__'):
                # システム予約の属性なのでフィールドではない
                continue
            # 取り出して連想配列に格納
            fc = getattr(cls.Field, k)
            from hatsudenki.packages.field import BaseHatsudenkiField
            if isinstance(fc, BaseHatsudenkiField):
                cls.__fields[k] = fc
                # ついでにフィールドに自分自身の名前を教えておく
                fc.name = k

    def __init__(self, name, parent: Markable, **kwargs):
        """
        イニシャライザ

        :param name: 名前
        :param parent: 親オブジェクト
        :param kwargs: メンバ初期化用
        """
        self._mark = set()
        self.name = name
        self._parent = parent

    def __setattr__(self, key, value):
        # 最初の設定ではない、且つフィールドを編集しようとしている
        if hasattr(self, key) and key in self.__class__.__fields:
            self.modify_mark(key)
        super().__setattr__(key, value)

    def modify_mark(self, key):
        """
        変更をマーク

        :param key: 変更したキー
        :return: None
        """
        self._mark.add(key)
        # 親オブジェクトが設定されていればマークを伝搬する
        if self._parent:
            self._parent.modify_mark(self.name)

    def flush(self):
        """
        変更情報をクリア

        :return: None
        """
        self._mark.clear()

    @classmethod
    def get_field_class(cls, name: str) -> SerializableField:
        """
        フィールド名から対応するフィールドクラスを取得

        :param name: フィールド名
        :return: フィールドクラス
        """
        return getattr(cls.Field, name)

    @classmethod
    def deserialize(cls, name, data: dict, parent: Markable):
        """
        DynamoDB → Python

        :param name: 名前
        :param data: dynamoDB dict
        :param parent: 親オブジェクト
        :return: 自分自身をインスタンス化したもの
        """
        a = cls(name=name, parent=parent)
        m = data['M']
        d = {}
        for k, f in cls.__fields.items():
            a.__setattr__(k, f.deserialize(m.get(k, None), a))
        return a

    @classmethod
    def serialize(cls, data: 'MarkedObject'):
        """
        Python → DynamoDB

        :param data: 変換対象
        :return: dynamoDB dict
        """
        ret = {
            'M': {k: v.serialize(getattr(data, k)) for k, v in data.__fields.items() if
                  not v.is_empty(getattr(data, k))}
        }
        return ret

    @property
    def serialized_value(self):
        """
        DynamoDB dictにシリアライズされた値

        :return: dynamoDB dict
        """
        return self.__class__.serialize(self)

    @classmethod
    def is_empty(cls, data):
        """
        値が空か判定

        :param data: 判定対象
        :return: boolean
        """
        return data is None

    def build_update_expression(self, parent_name, upd: UpdateExpression):
        """
        変更点を抽出してUpdateExpressionに追加する

        :param parent_name: 親の属性名
        :param upd: UpdateExpressionインスタンス
        :return: None
        """
        for update_key in self._mark:
            now_value = getattr(self, update_key)
            fc = self.__class__.get_field_class(update_key)
            n = parent_name + '.' + str(update_key)
            upd.set(n, fc.serialize(now_value), raw=True)



from pathlib import Path

from hatsudenki.packages.command.loader.base import LoaderBase
from hatsudenki.packages.command.master.enum.data import EnumData
from hatsudenki.packages.command.stdout.output import path_to_snake


class EnumLoader(LoaderBase[EnumData]):
    def __init__(self, base_path: Path):
        super().__init__(base_path)
        self.ext = '.yml'
        self.ref_table_name = {}

    def _load(self, path: Path):
        return EnumData(self.base_path, path)

    def setup(self, dir_name=''):
        super().setup(dir_name)

        def gen_key(p) -> str:
            return 'master_' + path_to_snake(str(Path(p).relative_to(self.base_path))).replace('.yml', '')

        self.ref_table_name = {gen_key(p): p for p, v in self.datas.items()}

    def get_by_name(self, enum_name: str):
        if enum_name.startswith('enum_'):
            enum_name = enum_name.replace('enum_', 'master_')
        if not enum_name.startswith('master_'):
            enum_name = 'master_' + enum_name
        try:
            self.ref_table_name[enum_name]
        except Exception as e:
            raise Exception(f'{enum_name} is not found.')
        return self.get_from_path(self.ref_table_name[enum_name])



from typing import Dict

from hatsudenki.packages.command.hatsudenki.field.base import HatsudenkiFieldBase
from hatsudenki.packages.command.hatsudenki.field.util import FieldFactory, field
from hatsudenki.packages.command.stdout.output import IndentString, snake_to_camel


class ChildObject(object):
    def __init__(self, data: dict):
        self.data = data
        # ファクトリを通してタイプに応じたフィールドクラスをインスタンス化した後に格納する
        self.attributes: Dict[str, HatsudenkiFieldBase] = {k: FieldFactory.create_field(k, v, self) for k, v in
                                                           self.data['attributes'].items()}

    def render(self, name):
        """
        クラス出力

        :param name: フィールド名
        :return: IndentString
        """
        u = IndentString()
        u.indent(f'class {name}(MarkedObject):')
        u.indent('class Field:')
        for k, v in self.attributes.items():
            u.add(v.class_string)
            u.add(v.gen_def_str())

        u.outdent()
        u.blank_line()
        u.indent('def __init__(self, name, parent: Markable, **kwargs):')
        u.add('super().__init__(name, parent)')
        u.add('ft = self.__class__.Field')
        for k, v in self.attributes.items():
            u.add(v.gen_init_str(direct_assign=False))

        u.add('')
        return u


class IndexedChildObject(ChildObject):

    def render(self, name):
        hash_key = next((v for k, v in self.attributes.items() if v.is_hash))

        u = IndentString()
        u.indent(f'class {name}(MarkedObjectWithIndex):')
        u.indent('class Meta:')
        u.add(f"hash_key = '{hash_key.name}'")
        u.outdent()
        u.blank_line()
        u.indent('class Field:')
        for k, v in self.attributes.items():
            u.add(v.class_string)
            u.add(v.gen_def_str())

        u.outdent()
        u.blank_line()
        u.indent('def __init__(self, name, parent: Markable, **kwargs):')
        u.add('super().__init__(name, parent)')
        u.add('ft = self.__class__.Field')
        for k, v in self.attributes.items():
            u.add(v.gen_init_str(direct_assign=False))

        u.add('')
        return u


@field('list', 'list')
class ListField(HatsudenkiFieldBase):
    PythonStr = 'field.ListField'

    @property
    def child(self):
        """
        小要素の定義

        :return: ChildObjectインスタンス
        """
        v = self.data.get('value')
        return ChildObject(v)

    def _parse_opt(self):
        f = super()._parse_opt()
        f.append(f'value_type={self.value_class_name}')
        return f

    @property
    def value_class_name(self):
        """
        小要素のタイプネーム

        :return: str
        """
        return snake_to_camel(self.name) + 'Value'

    @property
    def class_string(self):
        c = self.child
        return c.render(self.value_class_name)

    def gen_def_str(self, ex_opt: dict = None):
        return f"{self.name}: {self.PythonStr}[{self.value_class_name}] = {self.def_python_field_class(ex_opt)}"

    def gen_init_str(self, direct_assign=True):
        return f'self.{self.name} = ft.{self.name}.get_data_from_dict(kwargs, self)'


@field('dict', 'dict')
class DictField(HatsudenkiFieldBase):
    PythonStr = 'field.DictField'


@field('map', 'dict')
class MapField(HatsudenkiFieldBase):
    PythonStr = 'field.MapField'

    def _parse_opt(self):
        f = super()._parse_opt()
        f.append(f'value_type={self.value_class_name}')
        return f

    @property
    def value_class_name(self):
        """
        小要素のタイプネーム

        :return: str
        """
        return snake_to_camel(self.name) + 'Value'

    @property
    def class_string(self):
        c = self.child
        return c.render(self.value_class_name)

    @property
    def child(self):
        """
        小要素の定義

        :return: ChildObjectインスタンス
        """
        v = self.data
        return ChildObject(v)

    def gen_def_str(self, ex_opt: dict = None):
        return f"{self.name}: {self.PythonStr}[{self.value_class_name}] = {self.def_python_field_class(ex_opt)}"

    def gen_init_str(self, direct_assign=True):
        return f'self.{self.name} = ft.{self.name}.get_data_from_dict(kwargs, self)'


@field('dict_map', 'dict')
class DictMapField(ListField):
    PythonStr = 'field.DictMapField'

    @property
    def child(self):
        return IndexedChildObject(self.data.get('value'))



from collections import defaultdict
from pathlib import Path
from typing import Dict

from hatsudenki.packages.command.loader.base import LoaderBase
from hatsudenki.packages.command.master.enum.loader import EnumLoader
from hatsudenki.packages.command.master.table import MasterTable
from hatsudenki.packages.command.stdout.output import ToolOutput


class MasterTableLoader(LoaderBase[MasterTable]):
    """
    Master定義YAMLのローダー
    """

    def __init__(self, base_path: Path, enum_loader: EnumLoader):
        super().__init__(base_path)
        self.ext = '.yml'
        self.ref_table_name: Dict[str, MasterTable] = {}
        self.ref_excel_name: Dict[str, Dict[str, MasterTable]] = {}
        self.enum_loader = enum_loader

    def _load(self, path: Path):
        return MasterTable(self.base_path, path, self)

    def setup(self, dir_name=''):
        super().setup(dir_name)

        ToolOutput.anchor('MasterExcelを準備します')
        d = defaultdict(dict)
        for k, v in self.iter():
            self.ref_table_name[v.table_name] = v
            d[v.excel_name][v.excel_sheet_name] = v
        self.ref_excel_name = dict(d)
        ToolOutput.pop()

    def get_by_table_name(self, table_name: str):
        if not table_name.startswith('master_'):
            table_name = 'master_' + table_name
        return self.ref_table_name.get(table_name, None)

    def get_by_excel_name(self, excel_name: str):
        return self.ref_excel_name[excel_name]



from hatsudenki.define.config import TAG_SEPARATOR
from hatsudenki.packages.command.stdout.output import IndentString


class HatsudenkiFieldBase(object):
    """
    Fieldベース
    """
    # YAMLにて定義される値
    TypeStr = ''
    # アプリ側でのフィールドクラス名
    PythonStr = ''
    # アプリ側でのバリュークラス名
    PythonValueTypeStr = ''
    # 個別にクラスを生成する必要があるか
    HasGenClass = False

    def __init__(self, name: str, raw_dict: dict, parent):
        """
        イニシャライザ

        :param name: 名前
        :param raw_dict: yamlデータ
        :param parent: 親オブジェクト
        """
        from hatsudenki.packages.command.hatsudenki.data import HatsudenkiData
        self.name = name
        self.data = raw_dict
        self.parent: HatsudenkiData = parent

    @property
    def default_value(self):
        """
        デフォルト値

        :return: any
        """
        return self.data.get('default', None)

    @property
    def python_field_type_str(self):
        """
        Pythonフィールドクラス名
        :return: str
        """
        return self.__class__.PythonStr

    @property
    def python_value_type_str(self):
        """
        pythonバリュータイプ名

        :return: str
        """
        return self.__class__.PythonValueTypeStr

    @property
    def python_init_type_str(self):
        """
        イニシャライザ初期化パラメータとして受け取る型名
        :return: str
        """
        # 基本的にValueTypeと同一
        return self.__class__.PythonValueTypeStr

    def _parse_opt(self):
        """
        追加のオプション

        :return: list
        """
        # デフォルトではなにもない
        opt = []
        return opt

    def get_init_option(self, extra: dict = None):
        """
        イニシャライザのオプション
        :param extra: 追加オプション
        :return: str
        """
        # 型独自の追加オプションを処理
        opt = self._parse_opt()

        # デフォルト値があれば追加する
        d = self.default_value
        if d is not None:
            opt.append(f'default={d}')

        # 追加オプションがあれば追加する
        if extra is not None:
            for k, v in extra.items():
                opt.append(f'{k}={str(v)}')
        return ', '.join(opt)

    def gen_init_str(self, direct_assign=True):
        """
        イニシャライザ内での初期化処理

        :return: str
        """
        if direct_assign:
            if self.is_alias_key:
                return f'self.{self.name} = self.Meta.key_alias_type.get_data({self.name}, self)'
            elif self.is_hash or self.is_range:
                return f'self.{self.name}: {self.python_value_type_str} = ft.{self.name}.get_data({self.name}, self)'
        return f'self.{self.name}: {self.python_value_type_str} = ft.{self.name}.get_data_from_dict(kwargs, self)'

    def gen_def_str(self, ex_opt: dict = None):
        """
        Fieldクラス内での定義処理

        :param ex_opt: 追加オプション
        :return: str
        """
        return f"{self.name} = {self.def_python_field_class(ex_opt)}"

    def def_python_field_class(self, extra_opt: dict = None):
        """
        Fieldクラスの定義文字列

        :param extra_opt: 追加オプション
        :return: str
        """
        return f'{self.python_field_type_str}({self.get_init_option(extra_opt)})'

    @property
    def is_alias_key(self):
        """
        エイリアスキーか

        :return: boolean
        """
        return self.is_range and not self.parent.is_alone

    @property
    def is_hash(self):
        """
        ハッシュキーか

        :return: boolean
        """
        return self.data.get('hash', False)

    @property
    def is_range(self):
        """
        レンジキーか

        :return: boolean
        """
        return self.data.get('range', False)

    @property
    def resolver_string(self):
        """
        リゾルバ

        :return: IndexString
        """
        if not self.is_alias_key:
            # エイリアスキーでなければ不要
            return None

        # エイリアスキーはリゾルバが必要
        l = len(self.parent.tag_name) + len(TAG_SEPARATOR)
        parent = self.parent.parent_table
        body = IndentString()
        # getter
        body.add('@property')
        body.indent(f'def {self.name}(self) -> {self.PythonValueTypeStr}:')
        body.add(f'return {self.PythonValueTypeStr}(self.{parent.range_key.name}[{l}:])')
        body.outdent()

        body.blank_line()

        # setter
        body.add(f'@{self.name}.setter')
        body.indent(f'def {self.name}(self, val: {self.PythonValueTypeStr}):')
        body.indent('if val is None:')
        body.add('return')
        body.outdent()
        body.add(
            f'self.{self.parent.parent_table.range_key.name} = f"{self.parent.tag_name}{TAG_SEPARATOR}{self.resolver_setter_name}"')
        body.outdent()

        return body

    @property
    def resolver_setter_name(self):
        return '{val}'

    @property
    def class_string(self):
        return None



from pathlib import Path

from openpyxl.styles import Font, Color

from hatsudenki.packages.command.excel.excel_data import ExcelData, CellStyle
from hatsudenki.packages.command.master.enum.loader import EnumLoader
from hatsudenki.packages.command.master.loader import MasterTableLoader


class EnumExcelBook(ExcelData):

    def __init__(self, file_path: Path, enum_loader: EnumLoader, table_loader: MasterTableLoader):
        super().__init__(file_path)
        self.enum_loader = enum_loader
        self.table_loader = table_loader
        self.create()
        self._generate_sheets()

    def _generate_sheets(self):
        for k, data in self.enum_loader.iter():
            sheet_name = data.excel_sheet_name
            if sheet_name in self.book.sheetnames:
                self.book.remove(self.book[sheet_name])
            sheet = self.book.create_sheet(sheet_name)
            sheet.protection.enable()
            sheet.sheet_properties.tabColor = 'ff0000'

            if data.is_selector:
                self.write_row(sheet_name, 0, 0, ['id', 'name', 'ref_name', 'display_name', 'value', 'comment'],
                               CellStyle.RED)
                for idx, v in enumerate(data.values):
                    row = v.excel_row
                    val = row[4]
                    if val.startswith('enum_'):
                        target_enum = self.enum_loader.get_by_name(val)
                        l = f'=HYPERLINK("{target_enum.excel_name}.xlsx#{target_enum.excel_sheet_name}!A1", "【参照】{target_enum.excel_name} [{target_enum.excel_sheet_name}]")'
                    else:
                        target_table = self.table_loader.get_by_table_name(val)
                        l = f'=HYPERLINK("../{target_table.excel_name}.xlsx#{target_table.excel_sheet_name}!A1", "【参照】{target_table.excel_name} [{target_table.excel_sheet_name}]")'
                    row[4] = l
                    self.write_row(sheet_name, idx + 1, 0, row)
                    c = self.book[sheet_name].cell(idx + 2, 5)
                    c.font = Font(name="Meiryo UI", size=10, underline="single",
                                  color=Color(rgb=None, indexed=None, auto=None, theme=10, tint=0.0, type="theme"))

            else:
                self.write_row(sheet_name, 0, 0, ['id', 'name', 'ref_name', 'display_name', 'comment'], CellStyle.RED)
                for idx, v in enumerate(data.values):
                    self.write_row(sheet_name, idx + 1, 0, v.excel_row)



from pathlib import Path

from hatsudenki.packages.command.excel.excel_data import ExcelData


class MasterAssignerExcel(ExcelData):
    def __init__(self, file_path: Path):
        super().__init__(file_path)



from pathlib import Path

import yaml

from hatsudenki.commands.base import BaseCommand
from hatsudenki.packages.command.files import recreate_dir, write_file
from hatsudenki.packages.command.master.enum.loader import EnumLoader
from hatsudenki.packages.command.master.exporter.loader import MasterExcelLoader
from hatsudenki.packages.command.master.exporter.renderer.raw_yaml import RawYamlRenderer
from hatsudenki.packages.command.master.loader import MasterTableLoader
from hatsudenki.packages.command.master.tag.loader import MasterTagLoader, MasterTag
from hatsudenki.packages.command.stdout.output import ToolOutput


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument('-c', '--clean', action='store_true')
        parser.add_argument('-i', '--in', required=True)
        parser.add_argument('-o', '--out', required=True)
        parser.add_argument('-dsl', '--dsl_path', required=True)

        parser.add_argument('-t', '--tag', default='ALL')
        parser.add_argument('-d', '--debug', action='store_true')

    def clean(self, out_yaml_path: Path):
        recreate_dir(out_yaml_path)

    async def handle(self, *args, **options):
        # 引数解析
        in_path = Path(options.get('in'))
        out_path = Path(options.get('out'))
        dsl_path = Path(options.get('dsl_path'))
        target_tag: str = options.get('tag')
        enable_debug = options.get('debug', False)
        ToolOutput.print(f'dump master yaml... tag={target_tag} debug={enable_debug}', False)
        ToolOutput.out(f'parameter {in_path}, {out_path}, tag={target_tag}')

        # 出力TAG
        tag_loader = MasterTagLoader(dsl_path / 'define' / 'master_tag.yml')
        tag_loader.load()

        # TAG情報を取得
        out_tag = tag_loader.get_tag(target_tag)

        if out_tag is None:
            raise Exception(f'出力タグ {target_tag} は定義されていません！')

        if out_tag.is_only:
            ToolOutput.out(f'onlyモードで出力します include={out_tag.include}')

        enum_path = dsl_path / 'enum'
        master_schema_path = dsl_path / 'master'
        master_excel_path = in_path

        out_raw_yaml_path = out_path / 'raw_yaml'

        # Enum準備
        ToolOutput.print_with_anchor('enum準備')
        enum_loader = EnumLoader(enum_path)
        enum_loader.setup()
        ToolOutput.print_with_pop('OK')

        # スキーマパース
        ToolOutput.print_with_anchor('masterスキーマ準備')
        master_loader = MasterTableLoader(master_schema_path, enum_loader)
        master_loader.setup()
        ToolOutput.print_with_pop('OK')

        # Excel読み込み
        ToolOutput.print_with_anchor('excel準備')
        excel_loader = MasterExcelLoader(master_excel_path, master_loader, tag_loader)
        excel_loader.set_out_level(out_tag)
        excel_loader.set_debug_flg(enable_debug)
        excel_loader.setup()
        ToolOutput.print_with_pop('OK')

        # cleanフラグが設定されている場合は掃除する
        if options.get('clean', False):
            recreate_dir(out_raw_yaml_path)

        # YAML書き出し
        self.generate_yaml(excel_loader, out_raw_yaml_path, out_tag)

    def generate_yaml(self, excel_loader: MasterExcelLoader, out_raw_yaml_path: Path, out_tag: MasterTag):
        for p, excel in excel_loader.iter():
            r = RawYamlRenderer(excel)
            datas = r.render(out_tag)
            for k, v in datas.items():
                data = yaml.dump(v, allow_unicode=True, default_flow_style=False)
                write_file(out_raw_yaml_path / (k + '.yml'), data)



from pathlib import Path
from typing import Optional

from hatsudenki.packages.command.files import yaml_load
from hatsudenki.packages.command.stdout.output import snake_to_camel
from hatsudenki.packages.command.util.base_info import BaseInfo


class EnumValue(object):
    def __init__(self, val: int, data: dict):
        self.data = data
        self.id = val

    @property
    def name(self) -> str:
        return self.data['name']

    @property
    def enum_value_name(self):
        return self.name.upper()

    @property
    def excel_name(self):
        return self.data.get('ref_name', self.name)

    @property
    def display_name(self):
        return self.data.get('display_name', self.excel_name)

    @property
    def comment(self) -> str:
        return self.data.get('comment', '')

    @property
    def excel_row(self):
        return [self.id, self.name, self.excel_name, self.display_name, self.comment]

    @property
    def python_str(self):
        return f'{self.enum_value_name} = {self.id}'

    @property
    def cs_str(self):
        return f'{self.enum_value_name} = {self.id},'

    @property
    def cs_desc(self):
        return f'[Description("{self.display_name}")]'

    @property
    def ts_str(self):
        return f'{self.enum_value_name} = {self.id},'


class EnumValueDefine(EnumValue):
    pass


class EnumValueSelect(EnumValue):
    @property
    def excel_row(self):
        return [self.id, self.name, self.excel_name, self.display_name, self.select_value, self.comment]

    @property
    def select_value(self) -> Optional[str]:
        return self.data.get('value', None)

    @property
    def resolve_name(self):
        if self.select_value.startswith('enum_'):
            return f'if value == cls.{self.enum_value_name}: return EnumResolver({snake_to_camel(self.select_value)})'
        elif self.select_value.startswith('master_'):
            return f'if value == cls.{self.enum_value_name}: return MasterResolver(masters.{snake_to_camel(self.select_value)})'


class EnumData(BaseInfo):
    def __init__(self, base_path: Path, full_path: Path, *args, **kwargs):
        super().__init__(base_path, full_path, *args, **kwargs)
        self.data = yaml_load(full_path)

        tp = self.data.get('type', 'order')

        if tp == 'define':
            self.values = [EnumValueDefine(v['value'], v) for idx, v in enumerate(self.data['values'])]
        elif tp == 'selector':
            # selector
            self.origin = self.data.get('origin', 0)
            self.values = [EnumValueSelect(idx + self.origin, v) for idx, v in enumerate(self.data['values'])]
        else:
            self.origin = self.data.get('origin', 0)
            self.values = [EnumValue(idx + self.origin, v) for idx, v in enumerate(self.data['values'])]

    @property
    def label(self):
        return f'{self.excel_name}:{self.excel_sheet_name}:{self.rel_path}'

    @property
    def excel_name(self):
        return '定義タイプ'

    @property
    def excel_sheet_name(self):
        return self.data['name']

    @property
    def is_selector(self):
        return self.data.get('type', 'int') == 'selector'

    @property
    def is_out_desc(self) -> bool:
        return self.data.get('out_desc', False)

    def find(self, val, is_raw=False):
        # フォーマットの都合上必ず文字列で格納されているので、raw=trueの場合はintにキャストしてやる必要がある
        vv = int(val) if is_raw else val

        if is_raw:
            return next((v for v in self.values if v.id == vv), None)
        else:
            return next((v for v in self.values if v.excel_name == vv), None)



import unicodedata
from pathlib import Path
from typing import Dict, Optional

from hatsudenki.packages.command.loader.base import LoaderBase, T
from hatsudenki.packages.command.master.exporter.excel import MasterExcel
from hatsudenki.packages.command.master.loader import MasterTableLoader
from hatsudenki.packages.command.master.tag.loader import MasterTagLoader, MasterTag


class MasterExcelLoader(LoaderBase[MasterExcel]):
    """
    Masterデータ入力Excelローダー
    """

    def __init__(self, base_path: Path, master_loader: MasterTableLoader, master_tag_loader: MasterTagLoader):
        super().__init__(base_path)
        self.ext = '.xlsx'
        self.master_loader = master_loader
        self.ref_excel_name: Dict[str, MasterExcel] = {}
        self.out_tag: Optional[MasterTag] = None
        self.tag_loader = master_tag_loader
        self.is_debug = False

    def _load(self, path: Path) -> T:
        table = self.master_loader.get_by_excel_name(path.name.replace('.xlsx', ''))

        return MasterExcel(self.base_path, path, table, self)

    def set_out_level(self, out_tag: MasterTag):
        self.out_tag = out_tag

    def set_debug_flg(self, debug_flg: bool):
        self.is_debug = debug_flg

    def resolve_tag(self, tag_str: str):
        return self.tag_loader.get_tag(tag_str)

    def setup(self, dir_name=''):
        self.datas = {Path(unicodedata.normalize('NFC', str(path))): None for path in
                      (self.base_path / dir_name).glob('*' + self.ext) if not path.name.startswith('~$')}
        self.ref_excel_name = {p.name: ex for p, ex in self.iter()}

    def get_by_excel_name(self, name: str):
        """
        Excelファイル名を指定して取得

        :param name: 取得する対象のExcelファイル名
        :return:
        """
        return self.ref_excel_name[name]



from typing import Type, TypeVar, Iterator, Iterable

from hatsudenki.packages.table.index import PrimaryIndex

T = TypeVar('T')


class CacheBaseTableSolo(object):
    class Meta:
        table_name: str = None
        primary_index: PrimaryIndex = None

    _mapped = {}

    def __init__(self, **kwargs):
        pass

    @classmethod
    def _get(cls: Type[T], hash_key, range_key=None) -> T:
        """
        HASHキーを指定して一件取得

        :param hash_key: ハッシュキー値
        :param range_key: レンジキー値（Soloテーブルはこの値は必ずNoneになる）
        :return: 条件にマッチしたレコード
        """
        if range_key is not None:
            raise Exception('invalid key.')

        return cls._mapped[hash_key]

    @classmethod
    def _get_by_cursor(cls: Type[T], cursor: str) -> T:
        """
        カーソル文字列から一件取得

        :param cursor: カーソル文字列
        :return: 条件にマッチしたレコード
        """

        return cls._get(cursor)

    @classmethod
    def _check_exist_hash(cls, hash_key):
        """
        指定されたハッシュキーのレコードがあるか判定する

        :param hash_key: 検索対象値
        :return: bool
        """
        return hash_key in cls._mapped

    @classmethod
    def _iter(cls: Type[T]) -> Iterator[T]:
        """
        全件列挙イテレータ

        :return: iter
        """
        for d in cls._mapped.values():
            yield d

    @classmethod
    def get_hash_key_name(cls):
        """
        HASHキーの名前を取得する

        :return: ハッシュキー文字列
        """
        return cls.Meta.primary_index.hash_key

    @property
    def hash_value(self):
        """
        ハッシュキーの値を取得

        :return: ハッシュキーに設定された値
        """
        return getattr(self, self.get_hash_key_name())

    @classmethod
    def _set_records(cls, recs: Iterable[dict]):
        """
        レコードをセット

        :param recs: セットするレコード配列
        :return: None
        """

        hk = cls.get_hash_key_name()
        cls._mapped = {}
        for rec in recs:
            cls._mapped[rec[hk]] = cls(**rec)

    @property
    def one_cursor(self):
        """
        レコード一件を一意に表す文字列

        :return: str
        """
        return str(self.hash_value)

    def __str__(self) -> str:
        return self.one_cursor



from __future__ import annotations

import os
from collections import defaultdict
from os import PathLike
from pathlib import Path
from typing import Set, List, Dict

from hatsudenki.packages.command.files import yaml_load
from hatsudenki.packages.command.hatsudenki.field import HatsudenkiFieldBase, FieldFactory
from hatsudenki.packages.command.stdout.output import snake_to_camel


class HatsudenkiData(object):
    """
    Hatsudenkiデータスキーマ定義データ
    """

    class IndexInfo:
        """
        インデックス情報
        """

        def __init__(self):
            self.lsi: str = None
            self.gsi_hash: Set[str] = set()
            self.gsi_range: Set[str] = set()
            self.projection: Set[str] = set()

    def __init__(self, base_path: PathLike, full_path: PathLike, loader: any, *args, **kwargs):
        """
        イニシャライザ

        :param base_path: ベースパス
        :param full_path: フルパス
        :param loader: 自分を呼び出したローダ
        :param args:
        :param kwargs:
        """
        from .loader import HatsudenkiLoader
        self.full_path = Path(full_path)
        self.base_path = Path(base_path)
        self.data = yaml_load(self.full_path)
        # ファクトリを通してタイプに応じたフィールドクラスをインスタンス化した後に格納する
        self.attributes: Dict[str, HatsudenkiFieldBase] = {k: FieldFactory.create_field(k, v, self) for k, v in
                                                           self.data['attributes'].items()}

        self._table_name = None
        self._collection_name = None
        self._tag_name = None
        self.loader: HatsudenkiLoader = loader

    def _resolve_name(self):
        """
        テーブル名を解決

        :return: str
        """
        s = str(self.full_path.relative_to(self.base_path)).split('.')[0].replace(os.sep, '_')
        sp = s.split('_')
        if self.is_root:
            # rootテーブルはフォルダまでを基準とする
            sp.pop(-1)
        if len(sp) >= 2 and sp[-1] == sp[-2]:
            # 末尾2つの文字列が同一であった場合は省略する(user/user → user)
            sp = sp[:-1]
        return '_'.join(sp)

    @property
    def file_name(self):
        """
        ファイル名

        :return: str
        """
        return str(os.path.basename(self.full_path).split('.')[0])

    @property
    def is_root(self):
        """
        ルートテーブルか

        :return: boolean
        """
        return self.file_name == 'root'

    @property
    def is_single(self):
        """
        シングルテーブルか

        :return: boolean
        """
        return self.data.get('is_single', False)

    @property
    def is_alone(self):
        """
        ルートテーブル、もしくはシングルテーブルか

        :return: boolean
        """
        return self.is_root or self.is_single

    @property
    def label(self):
        """
        管理ツール用のテーブルかどうか
        :return:
        """
        return self.data.get('label', 'default')

    @property
    def parent_table(self):
        """
        親テーブルインスタンスを取得

        :return: 親テーブルインスタンス
        """
        if self.is_alone:
            # ルートテーブル、シングルテーブルには親が存在しない
            return None
        # 親を探す
        p = self.full_path.parent / 'root.yml'
        if self.loader.exists(p):
            return self.loader.get_from_path(p)

        # 存在しないエラー
        raise Exception('root not found.')

    @property
    def class_name(self):
        """
        ソース化したときのクラス名

        :return: str
        """
        return snake_to_camel(self.table_name)

    @property
    def collection_name(self):
        """
        コレクションネーム。シングルテーブルもしくはルートテーブルの場合はテーブルネームと同一。
        子テーブルの場合は親テーブルのテーブルネームとなる

        :return: str
        """
        if self._collection_name is None:
            if self.is_alone:
                # 単一テーブルの場合は自分の名前をそのまま返す
                self._collection_name = self.table_name
            else:
                # 子テーブルの場合は親の名前を返す
                self._collection_name = self.parent_table.collection_name
        return self._collection_name

    @property
    def table_name(self):
        """
        テーブル名

        :return: str
        """
        if self._table_name is None:
            self._table_name = self._resolve_name()
        return self._table_name

    @property
    def tag_name(self):
        """
        タグ名

        :return: str
        """
        if self.is_alone:
            # 単一テーブルはタグが存在しない
            return None

        if self._tag_name is None:
            # テーブル名から親テーブル名を削除したものをタグ文字列とする
            p = self.parent_table.table_name
            self._tag_name = self.table_name.replace(p + '_', '')
        return self._tag_name

    @property
    def indexes(self) -> List[dict]:
        """
        インデックス句

        :return: YAMLのIndex句
        """
        return self.data.get('indexes', [])

    @property
    def hash_key(self):
        """
        ハッシュキー

        :return: is_hash=Trueに設定されたattribute句の定義
        """
        return next((v for k, v in self.attributes.items() if v.is_hash), None)

    @property
    def range_key(self):
        """
        レンジキー

        :return: is_range=Trueに設定されたattribute句の定義
        """
        return next((v for k, v in self.attributes.items() if v.is_range), None)

    @property
    def capacity_units(self):
        """
        キャパシティユニット句

        :return: YAMLで定義されたcapacity_units句。省略された場合はリード、ライトともに1が入る。
        """
        return self.data.get('capacity_units', {'read': 1, 'write': 1})

    def parse_index(self):
        """
        インデックスを解釈し、適切なクラスをインスタンス化する。

        :return: IndexInfo, capacity_units
        """
        info = defaultdict(self.IndexInfo)
        units = {}

        for idx in self.indexes:
            # タイプ判定
            t = idx['type']
            if t == 'local':
                # LSIにはRANGEキーのみが設定されている。
                r = idx['range']
                l = f'_LSI-{r}'
                p = idx.get('projection', [])
                info[r].lsi = l
                [info[pk].projection.add(l) for pk in p]
            elif t == 'global':
                # GSIにはHASHとRANGEが存在し、キャパシティユニットが設定されている
                h = idx['hash']
                r = idx.get('range', '_')
                l = f'_GSI-{h}-{r}'
                p = idx.get('projection', [])
                info[h].gsi_hash.add(l)
                info[r].gsi_range.add(l)
                [info[pk].projection.add(l) for pk in p]
                units[l] = idx.get('capacity_units', {'read': 1, 'write': 1})
        return info, units



from typing import Optional


class PrimaryIndex(object):
    """
    インデックス情報
    """

    def __init__(self, hash_key: str, range_key: Optional[str] = None):
        """
        イニシャライザ

        :param hash_key: ハッシュキー名
        :param range_key: レンジキー名(ない場合はNone)
        """
        self.hash_key = hash_key
        self.range_key = range_key



from pathlib import Path

from hatsudenki.commands.base import BaseCommand
from hatsudenki.packages.command.files import write_file
from hatsudenki.packages.command.hatsudenki.loader import HatsudenkiLoader
from hatsudenki.packages.command.hatsudenki.renderer.tables import TableFileRenderer
from hatsudenki.packages.command.master.enum.loader import EnumLoader
from hatsudenki.packages.command.master.enum.renderer.excel import EnumExcelBook
from hatsudenki.packages.command.master.enum.renderer.python import PythonEnumRenderUnit, PythonEnumRenderer
from hatsudenki.packages.command.master.loader import MasterTableLoader
from hatsudenki.packages.command.master.renderer.excel import MasterExcelBook
from hatsudenki.packages.command.master.renderer.model import MasterModelRenderer
from hatsudenki.packages.command.stdout.output import ToolOutput


class Command(BaseCommand):

    def __init__(self, stdout=None, stderr=None, no_color=False):
        super().__init__(stdout, stderr, no_color)

    def add_arguments(self, parser):
        parser.add_argument('--dsl_path', '-dsl', type=str, required=True)
        parser.add_argument('--out', '-o', type=str, required=True)
        parser.add_argument('--clean', '-c', action='store_true')

    def clean(self, excel_path: Path, python_path: Path):
        ToolOutput.anchor('出力フォルダをクリア')
        ToolOutput.anchor('cleaning excel directory...')
        for book_file in excel_path.glob('*.xlsx'):
            ToolOutput.out(str(book_file))
            book_file.unlink()
        ToolOutput.pop('OK')

        ToolOutput.anchor('cleaning output python directory...')
        p = python_path / 'def_enum.py'
        if p.exists():
            p.unlink()
        p = python_path / 'masters.py'
        if p.exists():
            p.unlink()
        ToolOutput.pop('OK')
        ToolOutput.print_with_pop('OK')

    def load_enum_yaml(self, in_enum_path: Path):
        ToolOutput.anchor('loading enum yaml files...')
        enum_loader = EnumLoader(in_enum_path)
        enum_loader.setup()
        ToolOutput.pop('OK')
        return enum_loader

    def load_master_yaml(self, in_master_path: Path, enum_loader: EnumLoader):
        ToolOutput.anchor('loading master yaml files...')
        master_loader = MasterTableLoader(in_master_path, enum_loader)
        master_loader.setup()
        ToolOutput.pop('OK')
        return master_loader

    def render_dynamo(self, in_yaml_path: Path, out_python_path: Path):
        dynamo_loader = HatsudenkiLoader(in_yaml_path)
        dynamo_loader.setup()
        tbl_render = TableFileRenderer(dynamo_loader.iter(), str(out_python_path))
        write_file(out_python_path / 'tables.py', tbl_render.render())

    def render_master_excel(self, out_excel_path: Path, master_loader: MasterTableLoader):
        ToolOutput.anchor('generate excel books...')
        for excel_name, sheet_list in master_loader.ref_excel_name.items():
            ToolOutput.anchor(f'{excel_name}')
            book = MasterExcelBook(out_excel_path / (excel_name + '.xlsx'), sheet_list)
            ToolOutput.pop()
            book.save_file()
        ToolOutput.pop('OK')

    def render_enum_excel(self, out_excel_path: Path, enum_loader: EnumLoader, master_loader: MasterTableLoader):
        ToolOutput.anchor('generate enum excel...')
        enum_book = EnumExcelBook(out_excel_path / 'pg_excel' / '定義タイプ.xlsx', enum_loader, master_loader)
        enum_book.save_file()
        ToolOutput.pop('OK')

    def render_master_python(self, out_python_path: Path, master_loader: MasterTableLoader):
        ToolOutput.anchor('generate master python code...')
        model_renderer = MasterModelRenderer(master_loader)
        write_file(out_python_path / 'masters.py', model_renderer.render())
        ToolOutput.pop()

    def render_enum_python(self, out_python_path: Path, enum_loader: EnumLoader):
        ToolOutput.anchor('generate python enum code...')
        out_enum_path = out_python_path / 'def_enum.py'
        pr = PythonEnumRenderer()
        for p, d in enum_loader.iter():
            pr.add_unit(PythonEnumRenderUnit(d))
        write_file(out_enum_path, pr.render())
        ToolOutput.pop('OK')

    async def handle(self, *args, **options):
        ToolOutput.anchor('Generate Excel...')
        in_path = Path(options['dsl_path'])
        out_path = Path(options['out'])

        in_master_path = in_path / 'master'
        in_enum_path = in_path / 'enum'
        in_dynamo_path = in_path / 'dynamo'

        out_excel_path = out_path / 'excel'
        out_python_path = out_path / 'python'

        if options['clean']:
            self.clean(out_excel_path, out_python_path)

        enum_loader = self.load_enum_yaml(in_enum_path)
        master_loader = self.load_master_yaml(in_master_path, enum_loader)

        self.render_master_excel(out_excel_path, master_loader)

        # CSモデルの書き出し
        # out_cs_path = options.get('out_cs_path', None)
        # if out_cs_path:
        #     cs_model_path = Path(out_cs_path)
        #     ToolOutput.anchor('generate cs model...')
        #     for name, table in master_loader.iter():
        #         if not table.is_out_pack:
        #             continue
        #         p = cs_model_path / (table.class_name + '.cs')
        #         r = MasterCSModelRenderer()
        #         r.add_unit(MasterCSModelRenderUnit(table))
        #         write_file(p, r.render())
        #     ToolOutput.pop('OK')

        self.render_enum_excel(out_excel_path, enum_loader, master_loader)

        # pythonモデル＆データストアの書き出し
        self.render_master_python(out_python_path, master_loader)

        # generate python master models.
        self.render_enum_python(out_python_path, enum_loader)

        self.render_dynamo(in_dynamo_path, out_python_path)

        # generate cs master models.
        # if out_cs_path:
        #     ToolOutput.anchor('generate cs master models...')
        #     out_cs_enum_path = Path(out_cs_path) / 'enums'
        #     for p, d in enum_loader.iter():
        #         cs = CSEnumRenderer()
        #         cs.add_unit(CSEnumRenderUnit(d))
        #         write_file(out_cs_enum_path / ('Enum' + d.full_classname + '.cs'), cs.render())
        #     ToolOutput.pop('OK')
        ToolOutput.pop('all done!')



from enum import Enum
from typing import List, Type, TypeVar

from hatsudenki.packages.client import HatsudenkiClient
from hatsudenki.packages.table.solo import SoloHatsudenkiTable

MAX_TRANSACTION_ITEM = 10


class TransactionGetKind(Enum):
    Get = 'Get'


class QueryTransactGetItem(object):
    def __init__(self):
        self._task: List[dict] = []
        pass

    def _check(self):
        if len(self._task) >= MAX_TRANSACTION_ITEM:
            raise Exception(f'transaction item too many. max {MAX_TRANSACTION_ITEM}')

    def append(self, table: SoloHatsudenkiTable):
        self._check()
        query = {
            TransactionGetKind.Get.value: {
                'TableName': HatsudenkiClient.resolve_table_name(table.get_collection_name()),
                'Key': table.serialized_key
            }
        }

        self._task.append(query)

    async def exec(self):
        res = await HatsudenkiClient.transaction_get(self._task)
        return TransactGetResult(res)


T = TypeVar('T')


class TransactGetResult(object):
    def __init__(self, res: List[dict]):
        self._result = res

    def get(self, idx: int, target_type: Type[T]) -> T:
        return target_type.deserialize(self._result[idx]['Item'])



from collections import OrderedDict
from pathlib import Path
from typing import List, Dict, Generator

import openpyxl
from openpyxl.cell import Cell
from openpyxl.styles import NamedStyle, PatternFill, Font
from openpyxl.styles.colors import Color
from openpyxl.worksheet.worksheet import Worksheet


class CellStyle(object):
    NONE = NamedStyle(name='normal')
    BLUE = NamedStyle(name='fill_blue')
    BLUE.fill = PatternFill(fill_type='solid', fgColor='D9E1F2')
    RED = NamedStyle(name='fill_red')
    RED.fill = PatternFill(fill_type='solid', fgColor='FCE4D6')
    GREEN = NamedStyle(name='fill_green')
    GREEN.fill = PatternFill(fill_type='solid', fgColor='CCFFCC')


# HYPERLINK_REGEXP = re.compile('=HYPERLINK\(\"(([#!\.]?[^.]+)+)\", \"(([#!\.]?[^.]+)+)\"\)')


class ExcelData(object):
    def __init__(self, file_path: Path):
        self.file_path = file_path
        self.book: openpyxl.Workbook = None
        self.formula_book: openpyxl.Workbook = None

    @property
    def name(self):
        return self.file_path.name

    def load(self, flg=True):
        if self.book is None:
            self.book = openpyxl.load_workbook(str(self.file_path), data_only=flg)
        if flg:
            self.formula_book = openpyxl.load_workbook(str(self.file_path), data_only=False)

    def create(self):
        # ブック作成
        self.book = openpyxl.Workbook()
        self.book.remove(self.book['Sheet'])

    def get_row_values(self, sheet_name, offset_row, offset_col, width):
        sheet = self.book[sheet_name]
        return [sheet.cell(offset_row + 1, offset_col + i + 1).value for i in range(0, width)]

    def rect_value_iter(self, sheet_name: str, offset_row, offset_col, width):
        sheet: Worksheet = self.book[sheet_name]

        row_idx = 0
        while sheet.cell(row_idx + offset_row + 1, offset_col + 1).value is not None:
            yield self.get_row_values(sheet_name, row_idx + offset_row, offset_col, width)
            row_idx += 1

    def rect_value(self, sheet_name: str, offset_row, offset_col, width, all_str=False):
        ret: List[List[any]] = []
        row = offset_row + 1
        sheet: Worksheet = self.book[sheet_name]
        while sheet.cell(row, offset_col + 1).value is not None:
            if all_str:
                # 全て文字列に変換する
                vals = [str(sheet.cell(row, col + offset_col + 1).value) for col in range(0, width)]
            else:
                vals = [sheet.cell(row, col + offset_col + 1).value for col in range(0, width)]
            ret.append(vals)
            row += 1
        return ret

    def rect_value_to_dict(self, sheet_name, offset_row, offset_col, width):
        ret = []
        r = offset_row + 1
        sheet = self.book[sheet_name]

        # 一行目をヘッダとする
        header = [sheet.cell(r, x + offset_col + 1).value for x in range(width)]

        r += 1

        while sheet.cell(r, 1).value is not None:
            vals = {header[x]: sheet.cell(r, x + offset_col + 1).value for x in range(0, width)}
            ret.append(vals)
            r += 1
        return ret

    def get_sheet(self, sheet_name: str, formula: bool = False) -> Worksheet:
        if not formula or self.formula_book is None:
            return self.book[sheet_name]

        return self.formula_book[sheet_name]

    def iter_row(self, sheet_name: str) -> Generator[List[Cell], None, None]:
        sheet = self.get_sheet(sheet_name)
        return sheet.rows

    def iter_row_dict(self, sheet_name: str):
        sheet = self.get_sheet(sheet_name, formula=True)
        iter = sheet.rows
        h = next(iter, None)
        if h is None:
            return []

        key_names = []
        for cell in h:
            c = cell.value
            if c is None:
                key_names.append('column_' + str(len(key_names)))
            else:

                if c.startswith('=HYPERLINK'):
                    v1, v2 = c.split(',')
                    c = v2.strip()[1:-2]

                # m = HYPERLINK_REGEXP.match(c)
                # if m is not None:
                #     c = m.group(3)

                key_names.append(c)

        for cells in iter:
            v = [cell.value for cell in cells]
            if len(v) is 0 or v[0] is None:
                break
            l = zip(key_names, v)
            yield OrderedDict(l)

        # return (OrderedDict(zip(key_names, [cell.value for cell in cells])) for cells in iter)

    def max_column(self, sheet_name):
        sheet = self.book[sheet_name]
        return sheet.max_column

    def max_row(self, sheet_name):
        sheet = self.book[sheet_name]

        return sheet.max_row

    def add_named_style(self, style):
        if style.name in self.book.named_styles:
            return
        self.book.add_named_style(style)

    def write_row(self, sheet_name, row, col, data, style=CellStyle.NONE, header_links: List[str] = None,
                  column_size: Dict[str, int] = None):
        header_links: List[str] = header_links or []
        column_size: Dict[str, int] = column_size or {}
        self.add_named_style(style)
        i = 0
        sheet = self.get_sheet(sheet_name)
        for cell in (sheet.cell(row + 1, col + 1 + idx) for idx in range(len(data))):
            cell.value = data[i]
            cell.style = style.name
            column_size[cell.column] = max((column_size.get(cell.column, 0), len(str(cell.value))))

            if header_links is not None and len(header_links) > i and len(header_links[i]) > 0:
                cell.value = f'=HYPERLINK("{header_links[i]}", "{cell.value}")'
                cell.font = Font(name="Meiryo UI", size=10, underline="single",
                                 color=Color(rgb=None, indexed=None, auto=None, theme=10, tint=0.0, type="theme"))

            i += 1

    def insert_row(self, sheet_name, row, col, data, style=CellStyle.NONE):
        sheet = self.get_sheet(sheet_name)
        sheet.insert_rows(row + 1, 1)
        self.write_row(sheet_name, row, col, data, style)

    def save_file(self):
        if len(self.book.worksheets) is 0:
            return
        self.file_path.parent.mkdir(parents=True, exist_ok=True)
        self.book.save(str(self.file_path))

    @classmethod
    def init_column_size(cls, st: Worksheet, column_size: Dict[str, int]):
        for col, value in column_size.items():
            st.column_dimensions[col].width = max(value * 2, 10)

    def get_header_values(self, sheet_name):
        book = self.formula_book or self.book

        first_line = next(book[sheet_name].rows)
        ret = {}
        for cell in first_line:
            c = cell.value
            if c is None:
                continue
            c = str(c)
            if c.startswith('=HYPERLINK'):
                v1, v2 = c.split(',')
                c = v2.strip()[1:-2]

            ret[c] = cell.coordinate
        return ret



from hatsudenki.packages.command.master.enum.data import EnumData
from hatsudenki.packages.command.renderer.base import RenderUnit, FileRenderer
from hatsudenki.packages.command.stdout.output import IndentString


class PythonEnumRenderUnit(RenderUnit[EnumData]):

    def _enum(self):
        u = IndentString()
        u.indent(f'class Enum{self.data.full_classname}(enum.IntEnum):')
        for v in self.data.values:
            u.add(v.python_str)

        if self.data.is_selector:
            u.blank_line()
            u.add('@classmethod')
            u.indent('def resolve(cls, value: int):')
            u.add('from common import masters')
            for v in self.data.values:
                u.add(v.resolve_name)
            u.outdent()

        # 日本語名を出力します
        u.blank_line()
        u.add('@classmethod')
        u.indent('def reverse(cls, value: int) -> str:')
        for v in self.data.values:
            u.add(f'if value == cls.{v.enum_value_name}: return "{v.excel_name}"')
        u.outdent()

        u.blank_line(2)

        return u

    def _resolver(self):
        pass

    def render(self) -> IndentString:
        en = self._enum()
        return en


class PythonEnumRenderer(FileRenderer):
    def render_header(self):
        u = IndentString()
        u.add('import enum')
        u.add('from hatsudenki.packages.master.model import EnumResolver, MasterResolver')
        u.blank_line(2)
        return u



from __future__ import annotations

from pathlib import Path
from typing import Tuple

import yaml

from hatsudenki.packages.command.master.compare.schema import Schema


class YMLSchemaReader(object):
    def __init__(self, base_path: Path):
        self.base_path = base_path
        self._data = []

    def setup(self) -> YMLSchemaReader:
        print(f'loading...yml files...')
        yml_files = [open(p) for p in self.base_path.glob('**/*.yml')]
        self._data = [yaml.safe_load(f) for f in yml_files]
        [f.close() for f in yml_files]
        return self

    def schema(self) -> Schema:
        # 初期の空スキーマを作成する
        groups = {f['excel'] for f in self._data}
        schema = Schema([Schema.Group(g, []) for g in groups])
        # 割当
        for k, tbl in (self._create_table(d) for d in self._data):
            next(g for g in schema.groups if g.name == k).tables.append(tbl)
        return schema

    def _create_table(self, dic: dict) -> Tuple[str, Schema.Group.Table]:
        print(f'    loading... {dic["name"]}')
        columns = [self._create_column(dic['column'][d]) for d in dic['column']]
        return dic['excel'], Schema.Group.Table(dic['name'], columns)

    @staticmethod
    def _create_column(dic: dict) -> Schema.Group.Table.Column:
        return Schema.Group.Table.Column(dic.get('name', dic.get('to')))



from dataclasses import dataclass
from enum import Enum
from typing import List

from hatsudenki.packages.client import HatsudenkiClient
from hatsudenki.packages.expression.update import UpdateExpression
from hatsudenki.packages.table.solo import SoloHatsudenkiTable


class TransactionWriteKind(Enum):
    Put = 'Put'
    Delete = 'Delete'
    Update = 'Update'


@dataclass
class TransactWriteTask:
    table_name: str = None
    item: dict = None
    condition: str = None


MAX_TRANSACTION_ITEM = 10


class QueryTransactWriteItem(object):
    def __init__(self):
        self._task: List[dict] = []

    def _check(self):
        if len(self._task) >= MAX_TRANSACTION_ITEM:
            raise Exception(f'transaction item too many. max {MAX_TRANSACTION_ITEM}')

    def append_put(self, table: SoloHatsudenkiTable, overwrite=False):
        self._check()
        cond = table.not_exist_condition() if not overwrite else None

        query = {
            TransactionWriteKind.Put.value: {
                'TableName': HatsudenkiClient.resolve_table_name(table.get_collection_name()),
                'Item': table.serialize(),

                **(cond.to_parameter() if cond is not None else {})
            }
        }

        self._task.append(query)

    def append_delete(self, table: SoloHatsudenkiTable):
        self._check()
        query = {
            TransactionWriteKind.Delete.value: {
                'TableName': HatsudenkiClient.resolve_table_name(table.get_collection_name()),
                'Key': table.serialized_key,
            }
        }

        self._task.append(query)

    def append_update(self, table: SoloHatsudenkiTable):
        self._check()
        upd = UpdateExpression()

        table.build_update_expression(upd)
        cond = table.exist_condition()

        upd.add('_v', 1)
        cond.op_and()
        with cond:
            cond.equal('_v', table._v)
            cond.op_or()
            cond.attribute_exists('_v')
        key = table.serialized_key

        query = {
            TransactionWriteKind.Update.value: {
                'TableName': HatsudenkiClient.resolve_table_name(table.get_collection_name()),
                'Key': key,
                **(upd.to_parameter(cond))
            }
        }

        self._task.append(query)

    async def exec(self):
        res = await HatsudenkiClient.transaction_write(self._task)
        return res



from collections import defaultdict, Counter
from dataclasses import dataclass, field
from pprint import pprint
from typing import Dict, DefaultDict, Counter as Counter_type

QUERY_COUNTER_LABEL_DEFINE = [
    'get',
    'put',
    'update',
    'query',
    'batch_write',
    'batch_get',
    'transact_write',
    'transact_get'
]


@dataclass
class QueryCounterUnit:
    counters: Dict[str, DefaultDict[str, int]] = field(default_factory=dict)
    total_counter: Counter_type = field(default_factory=Counter)
    read_ccu_counter: Counter_type = field(default_factory=Counter)
    read_ccu_detail_dict: Dict = field(default_factory=dict)
    total_read_ccu: float = 0
    write_ccu_counter: Counter_type = field(default_factory=Counter)
    write_ccu_detail_dict: Dict = field(default_factory=dict)
    total_write_ccu: float = 0

    def count(self, label, key='unknown'):
        counter = self.counters.get(label, None)
        if counter is None:
            self.counters[label] = defaultdict(int)
            counter = self.counters[label]
        counter[key] += 1
        self.total_counter[label] += 1

    def count_read_ccu(self, info: Dict):
        self.read_ccu_counter[info['TableName']] += info['CapacityUnits']
        self.total_read_ccu += info['CapacityUnits']
        b_target = self.read_ccu_detail_dict.get(info['TableName'], False)
        if b_target is False:
            self.read_ccu_detail_dict[info['TableName']] = defaultdict(int)
        self.read_ccu_detail_dict[info['TableName']]['Table'] += info['Table']['CapacityUnits']
        gsi_info = info.get('GlobalSecondaryIndexes', False)
        if gsi_info is False:
            return
        for key, dic in info['GlobalSecondaryIndexes'].items():
            self.read_ccu_detail_dict[info['TableName']][key] += dic['CapacityUnits']

    def count_write_ccu(self, info: Dict):
        self.write_ccu_counter[info['TableName']] += info['CapacityUnits']
        self.total_write_ccu += info['CapacityUnits']
        b_target = self.write_ccu_detail_dict.get(info['TableName'], False)
        if b_target is False:
            self.write_ccu_detail_dict[info['TableName']] = defaultdict(int)
        self.write_ccu_detail_dict[info['TableName']]['Table'] += info['Table']['CapacityUnits']
        gsi_info = info.get('GlobalSecondaryIndexes', False)
        if gsi_info is False:
            return
        for key, dic in info['GlobalSecondaryIndexes'].items():
            self.write_ccu_detail_dict[info['TableName']][key] += dic['CapacityUnits']

    def init_counter(self):
        self.counters = dict()
        self.total_counter = Counter()
        self.read_ccu_counter = Counter()
        self.read_ccu_detail_dict = dict()
        self.total_read_ccu = 0
        self.write_ccu_counter = Counter()
        self.write_ccu_detail_dict = dict()
        self.total_write_ccu = 0


class QueryCounter:
    _counter: QueryCounterUnit = None

    @classmethod
    def _get_context_data(cls):
        if cls._counter is None:
            cls._counter = QueryCounterUnit()
        return cls._counter

    @classmethod
    def count(cls, label, key='unknown'):
        query_counter = cls._get_context_data()
        query_counter.count(label, key)

    @classmethod
    def count_read_ccu(cls, info: Dict):
        query_counter = cls._get_context_data()
        query_counter.count_read_ccu(info)

    @classmethod
    def count_write_ccu(cls, info: Dict):
        query_counter = cls._get_context_data()
        query_counter.count_write_ccu(info)

    @classmethod
    def init_counter(cls):
        query_counter = cls._get_context_data()
        query_counter.init_counter()

    @classmethod
    def dump_counter(cls):
        query_counter = cls._get_context_data()
        pprint(query_counter.counters)
        pprint(query_counter.total_counter)
        print('-read consume cu-')
        pprint(query_counter.read_ccu_counter)
        print(f'total:{query_counter.total_read_ccu}')
        print('-write consume cu-')
        pprint(query_counter.write_ccu_counter)
        print(f'total:{query_counter.total_write_ccu}')
        print('-read detail-')
        pprint(query_counter.read_ccu_detail_dict)
        print('-write detail-')
        pprint(query_counter.write_ccu_detail_dict)

    @classmethod
    def get_consumed_cu(cls):
        query_counter = cls._get_context_data()
        return query_counter.total_read_ccu, query_counter.total_write_ccu

    @classmethod
    def get_cu_detail_counters(cls):
        query_counter = cls._get_context_data()
        return query_counter.read_ccu_detail_dict, query_counter.write_ccu_detail_dict
